from pathlib import Path
from typing import DefaultDict, Dict, Optional, Union

from openghg.store.base import BaseStore
from xarray import Dataset

__all__ = ["EulerianModel"]


# TODO: Currently built around these keys but will probably need more unique distiguishers for different setups
# model name
# species
# date (start_date)
# ...
# setup (included as option for now)


class EulerianModel(BaseStore):
    """This class is used to process Eulerian model data"""

    _root = "EulerianModel"
    _uuid = "63ff2365-3ba2-452a-a53d-110140805d06"
    _metakey = f"{_root}/uuid/{_uuid}/metastore"

    @staticmethod
    def read_file(
        filepath: Union[str, Path],
        model: str,
        species: str,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        setup: Optional[str] = None,
        overwrite: bool = False,
    ) -> Dict:
        """Read Eulerian model output

        Args:
            filepath: Path of Eulerian model species output
            model: Eulerian model name
            species: Species name
            start_date: Start date (inclusive) associated with model run
            end_date: End date (exclusive) associated with model run
            setup: Additional setup details for run
            overwrite: Should this data overwrite currently stored data.
        """
        # TODO: As written, this currently includes some light assumptions that we're dealing with GEOSChem SpeciesConc format.
        # May need to split out into multiple modules (like with ObsSurface) or into separate retrieve functions as needed.

        from collections import defaultdict

        from openghg.store import assign_data, datasource_lookup, load_metastore
        from openghg.util import clean_string, hash_file, timestamp_now, timestamp_tzaware
        from pandas import Timestamp as pd_Timestamp
        from xarray import open_dataset

        model = clean_string(model)
        species = clean_string(species)
        start_date = clean_string(start_date)
        end_date = clean_string(end_date)
        setup = clean_string(setup)

        filepath = Path(filepath)

        em_store = EulerianModel.load()
        metastore = load_metastore(key=em_store._metakey)

        file_hash = hash_file(filepath=filepath)
        if file_hash in em_store._file_hashes and not overwrite:
            raise ValueError(
                f"This file has been uploaded previously with the filename : {em_store._file_hashes[file_hash]}."
            )

        em_data = open_dataset(filepath)

        # Check necessary 4D coordinates are present and rename if necessary (for consistency)
        check_coords = {
            "time": ["time"],
            "lat": ["lat", "latitude"],
            "lon": ["lon", "longitude"],
            "lev": ["lev", "level", "layer", "sigma_level"],
        }
        for name, coord_options in check_coords.items():
            for coord in coord_options:
                if coord in em_data.coords:
                    break
            else:
                raise ValueError("Input data must contain one of '{coord_options}' co-ordinate")
            if name != coord:
                print("Renaming co-ordinate '{coord}' to '{name}'")
                em_data = em_data.rename({coord: name})

        attrs = em_data.attrs

        # author_name = "OpenGHG Cloud"
        # em_data.attrs["author"] = author_name

        metadata = {}
        metadata.update(attrs)

        metadata["model"] = model
        metadata["species"] = species
        metadata["processed"] = str(timestamp_now())

        if start_date is None:
            if len(em_data["time"]) > 1:
                start_date = str(timestamp_tzaware(em_data.time[0].values))
            else:
                try:
                    start_date = attrs["simulation_start_date_and_time"]
                except KeyError:
                    raise Exception("Unable to derive start_date from data, please provide as an input.")
                else:
                    start_date = timestamp_tzaware(start_date)
                    start_date = str(start_date)

        if end_date is None:
            if len(em_data["time"]) > 1:
                end_date = str(timestamp_tzaware(em_data.time[-1].values))
            else:
                try:
                    end_date = attrs["simulation_end_date_and_time"]
                except KeyError:
                    raise Exception("Unable to derive `end_date` from data, please provide as an input.")
                else:
                    end_date = timestamp_tzaware(end_date)
                    end_date = str(end_date)

        date = str(pd_Timestamp(start_date).date())

        metadata["date"] = date
        metadata["start_date"] = start_date
        metadata["end_date"] = end_date

        metadata["max_longitude"] = round(float(em_data["lon"].max()), 5)
        metadata["min_longitude"] = round(float(em_data["lon"].min()), 5)
        metadata["max_latitude"] = round(float(em_data["lat"].max()), 5)
        metadata["min_latitude"] = round(float(em_data["lat"].min()), 5)

        history = metadata.get("history")
        if history is None:
            history = ""
        metadata["history"] = history + f" {str(timestamp_now())} Processed onto OpenGHG cloud"

        key = "_".join((model, species, date))

        model_data: DefaultDict[str, Dict[str, Union[Dict, Dataset]]] = defaultdict(dict)
        model_data[key]["data"] = em_data
        model_data[key]["metadata"] = metadata

        required = ("model", "species", "date")
        lookup_results = datasource_lookup(metastore=metastore, data=model_data, required_keys=required)

        data_type = "eulerian_model"
        datasource_uuids = assign_data(
            data_dict=model_data,
            lookup_results=lookup_results,
            overwrite=overwrite,
            data_type=data_type,
        )

        em_store.add_datasources(uuids=datasource_uuids, data=model_data, metastore=metastore)

        # Record the file hash in case we see this file again
        em_store._file_hashes[file_hash] = filepath.name

        em_store.save()
        metastore.close()

        return datasource_uuids



from pathlib import Path
from typing import Dict, Optional, Union, cast

import xarray as xr


def parse_openghg(
    data_filepath: Union[str, Path],
    site: Optional[str] = None,
    species: Optional[str] = None,
    network: Optional[str] = None,
    inlet: Optional[str] = None,
    instrument: Optional[str] = None,
    sampling_period: Optional[str] = None,
    calibration_scale: Optional[str] = None,
    data_owner: Optional[str] = None,
    data_owner_email: Optional[str] = None,
    **kwargs: str,
) -> Dict:
    """
    Parse and extract data from pre-formatted netcdf file which already
    matches expected OpenGHG format.

    At the moment this must also be for a site known to OpenGHG. See
    'acrg_site_info.json' file.

    The arguments specified below are the metadata needed to store this
    surface observation file within the object store. If these keywords are
    not included within the attributes of the netcdf file being passed then
    these arguments must be specified.

    Args:
        site: Site code/name e.g. "TAC" (for Tacolneston)
        species: Species name or synonym e.g. "ch4"
        network: Network name. e.g. "DECC"
        inlet: Height of inlet for observation. e.g. "10m"
        instrument: Instrument name used for measurement e.g. "gcmd".
            Can be set to "NA" if this is an unknown detail.
        sampling_period: Sampling period in pandas style
            (e.g. 2H for 2 hour period, 2m for 2 minute period).
        calibration_scale: Calibration scale used for measurements
            e.g. "WMOX2007"
        data_owner: Name of data owner.
        data_owner_email: Email address for data owner.
        kwargs: Any additional attributes to be associated with the data.

    Returns:
        Dict : Dictionary of source_name : data, metadata, attributes
    """
    from openghg.standardise.meta import assign_attributes, metadata_default_keys
    from openghg.util import clean_string, load_json, synonyms

    data_filepath = Path(data_filepath)

    if data_filepath.suffix != ".nc":
        raise ValueError("Input file must be a .nc (netcdf) file.")

    data = xr.open_dataset(data_filepath)  # Change this to with statement?

    # Extract current attributes from input data
    attributes = data.attrs

    # Define metadata based on input arguments.
    metadata_initial = {
        "site": site,
        "species": species,
        "network": network,
        "inlet": inlet,
        "instrument": instrument,
        "sampling_period": sampling_period,
        "calibration_scale": calibration_scale,
        "data_owner": data_owner,
        "data_owner_email": data_owner_email,
    }

    # TODO: Decide if to allow any of these to be missed.

    # Populate metadata with values from attributes if inputs have not been passed
    for key, value in metadata_initial.items():
        if value is None:
            try:
                metadata_initial[key] = attributes[key]
            except KeyError:
                raise ValueError(f"Input '{key}' must be specified if not included in data attributes.")
        else:
            # If attributes are present, check these match to inputs passed
            if key in attributes:
                attributes_value = attributes[key]
                if value != attributes_value:
                    # If inputs do not match attribute values, raise a ValueError
                    raise ValueError(
                        f"Input for '{key}': {value} does not match value in file attributes: {attributes_value}"
                    )

    metadata = cast(Dict[str, str], metadata_initial)

    metadata["inlet_height_magl"] = metadata["inlet"]

    # Define remaining keys needed for metadata
    metadata_needed = metadata_default_keys()
    metadata_needed = [param for param in metadata_needed if param not in metadata]

    metadata["site"] = clean_string(metadata["site"])
    metadata["species"] = synonyms(metadata["species"]).lower()
    # May want to remove the .lower() here and centralise this

    # Update attributes to match metadata after cleaning
    attributes["site"] = metadata["site"]
    attributes["species"] = metadata["species"]

    site = metadata["site"]
    network = metadata["network"]
    species = metadata["species"]

    # Allow site and network data to be treated in a case insensitive way
    site_case_options = [site, site.upper(), site.lower()]
    network_case_options = [network, network.upper(), network.lower()]

    # Extract centralised data for site (if present)
    site_data = load_json(filename="acrg_site_info.json")
    for site_value in site_case_options:
        if site_value in site_data:
            site_info_all = site_data[site_value]
            break
    else:
        print("Unknown site. Will attempt to extract metadata from dataset attributes or input keywords")
        site_info_all = {}

    for network_value in network_case_options:
        if network_value in site_info_all:
            site_info = site_info_all[network_value]
            break
    else:
        print(
            "Network {network} does not match with site {site}. Will attempt to extract metadata from dataset attributes or input keywords"
        )
        site_info = {}

    if site_info:
        # Ensure keywords match to metadata names for station values
        # e.g. "station_longitude" derived from "longitude"
        for key in metadata_needed:
            prefix = "station_"
            if key.startswith(prefix):
                split_key = key.split("_")[1:]
                short_key_option1 = "_".join(split_key)

                split_key.insert(1, "station")  # to catch "height_station_masl"
                short_key_option2 = "_".join(split_key)

                short_key_options = [short_key_option1, short_key_option2]
                for short_key in short_key_options:
                    if short_key in site_info:
                        site_info[key] = site_info[short_key]
                        break

    # Load attributes data for network if present
    param_data = load_json(filename="attributes.json")
    for network_value in network_case_options:
        if network_value in param_data:
            network_params = param_data[network_value]
            site_attributes = network_params["global_attributes"]
            break
    else:
        site_attributes = {}

    # Define sources of attributes to use when defining metadata
    # The order here influences the hierarchy if keys appear multiple times.
    # kwargs allow additional variables such as "station_longitude" to be included if needed.
    attribute_sources = [attributes, kwargs, site_info, site_attributes]

    # Search attributes sources (in order) and populate metadata
    for param in metadata_needed:
        for source in attribute_sources:
            if param in source:
                metadata[param] = source[param]
                break
        else:
            raise ValueError(
                f"Cannot extract or infer '{param}' parameter needed for metadata from stored data, attributes or keywords"
            )

    gas_data = {species: {"metadata": metadata, "data": data, "attributes": attributes}}

    gas_data = assign_attributes(data=gas_data, site=site, network=network)

    return gas_data



"""
Call OpenGHG serverless functions
"""
import os
from pathlib import Path
from typing import Dict, Optional, Tuple

import msgpack
import requests
from openghg.types import FunctionError
from openghg.util import compress, hash_bytes


def create_file_package(filepath: Path, obs_type: str) -> Tuple[bytes, Dict]:
    """Reads file metadata and compresses data to be sent to the serverless function

    Args:
        filepath: Path of data
        obs_type: Observation type
    Returns:
        tuple: Compressed data as bytes and dictionary of file metadata
    """
    in_mem_limit = 300  # MB
    # To convert bytes to megabytes
    MB = 1e6

    # Get the file size in megabytes
    file_size = filepath.stat().st_size / MB

    if file_size > in_mem_limit:
        raise NotImplementedError("We can't handle this size of file yet.")

    file_data = filepath.read_bytes()
    sha1_hash = hash_bytes(data=file_data)
    compressed_data = compress(data=file_data)
    filename = filepath.name

    file_metadata = {
        "sha1_hash": sha1_hash,
        "filename": filename,
        "compressed": True,
        "obs_type": obs_type.lower(),
    }

    return compressed_data, file_metadata


def create_post_dict(
    function_name: str,
    data: bytes,
    metadata: Dict,
    file_metadata: Dict,
    precision_data: Optional[Dict] = None,
    precision_file_metadata: Optional[Dict] = None,
) -> Dict:
    """Create the dictionary to POST to the remote function

    Args:
        function: Function name
        data: Compressed data
        metadata: Metadata dictionary
        file_metadata: File metadata
        precision_data: GCWERKS precision data
        precision_file_metadata: GCWERKS precision file metadata
    Returns:
        dict: Dictionary ready for function call
    """
    to_post = {
        "function": function_name,
        "data": data,
        "metadata": metadata,
        "file_metadata": file_metadata,
    }

    if precision_data is not None:
        to_post["precision_data"] = precision_data

        if precision_file_metadata is None:
            raise ValueError("We need both precision data and file metadata.")

        to_post["precision_file_metadata"] = precision_file_metadata

    return to_post


def call_function(data: Dict) -> Dict:
    """Calls an OpenGHG serverless function and returns its response

    Args:
        data: Data to POST. Must be a dictionary created using the create_post_dict function.
    Returns:
        dict: Dictionary containing response status, headers and content.
    """
    fn_url = _get_function_url()
    auth_key = _get_auth_key()

    headers = {}
    headers["Content-Type"] = "application/octet-stream"
    headers["authorization"] = auth_key

    packed_data = msgpack.packb(data)
    response = requests.post(url=fn_url, data=packed_data, headers=headers)

    if response.status_code != 200:
        raise FunctionError(f"Function call error: {str(response.content)}")

    return {
        "status": response.status_code,
        "headers": dict(headers),
        "content": msgpack.unpackb(response.content),
    }


def _get_function_url() -> str:
    """Get the URl for the required service

    Args:
        service_name: Service name
    Returns:
        str: Service / function URL
    """
    try:
        return os.environ["OPENGHG_FN_URL"]
    except KeyError:
        raise FunctionError("No OPENGHG_FN_URL environment variable set for function URLs")


def _get_auth_key() -> str:
    """Get the authentication key from the local environmen

    This offers very limited control over calling of the functions for now.
    It will be replaced by whatever user authentication system we end up using.

    Returns:
        str: Authentication key for serverless Fn
    """
    try:
        return os.environ["AUTH_KEY"]
    except KeyError:
        raise FunctionError("A valid AUTH_KEY secret must be set.")



__all__ = ["call_function", "create_file_package", "create_post_dict"]

from ._call import call_function, create_file_package, create_post_dict
from ._packaging import package_from_function, unpackage



from pathlib import Path

import paramiko

# See this for tests
# https://pypi.org/project/mock-ssh-server/

# type: ignore


class SSHConnect:
    """Use Paramiko to connect to an SSH server

    This class can be used as a context manager
    """

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __init__(self):
        self._host_keys = {}
        self._port = 22
        self._hostkey = None
        # This will be the paramiko.SSHClient object
        self._client = None

    def close(self):
        """Calls close on the paramiko SSHClient instance used
        the SSH connection

        Returns:
            None
        """
        try:
            self._client.close()
        except AttributeError:
            return

    def connect(self, username, hostname, keypath=None, password=None, known_host=False):
        """Use Paramiko to connect the hostname

        Args:
            user (str): Username for login
            hostname (str): Hostname of server. This can also be an IP address
            key_path (Path or str, default=None): Path to private key to use for authentication.
            By default Paramiko will attempt to use any “id_rsa”, “id_dsa” or “id_ecdsa” key
            discoverable in ~/.ssh/
            passsword (str, default=None): Password used for authentication or for unlocking a
            password protected
            private key
        Returns:
            None
        """
        self._client = paramiko.SSHClient()
        self._client.load_system_host_keys()

        # If we haven't connected to the host before the host's fingerprint
        # will not be in the known_hosts file this lets us override the fingerprint checking
        if not known_host:
            self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        self._client.connect(
            hostname=hostname,
            port=22,
            username=username,
            key_filename=keypath,
            password=password,
        )

    def run_command(self, commands):
        """Run commands on the remote server

        Args:
            commands (str, list): Command(s) to be run on the remote server
        Returns:
            list: List of tuples of stdin, stdout and stderr for each command
        """
        responses = {}

        if not isinstance(commands, list):
            commands = [commands]

        for c in commands:
            stdin, stdout, stderr = self._client.exec_command(c)

            # Get bytes from the objects and convert to str
            stdout = stdout.read().decode("utf-8")
            stderr = stderr.read().decode("utf-8")

            responses[c] = {"stdout": stdout, "stderr": stderr}

        return responses

    def write_files(self, files, remote_dir=None):
        """Write the job script to the remote server

        Args:
            files (list): List of paths of files to transfer
            path (str, default=None): Path to write script, if not passed the script
            will be written to the user's home directory
        Returns:
            None
        """
        sftp = self._client.open_sftp()

        if not isinstance(files, list):
            files = [files]

        # Convert to pathlib.Path objects for easier handling
        files = [Path(f) for f in files]

        if remote_dir is not None:
            remote_dir = Path(remote_dir)
            try:
                sftp.mkdir(str(remote_dir))
            except IOError:
                # This probably means the directory already exists
                # TODO - better way of handling this error?
                pass

        for filepath in files:
            # Here we only want the filename
            if remote_dir is not None:
                remote_path = str(remote_dir.joinpath(filepath.name))
            else:
                remote_path = filepath.name

            # Hopefully paramiko will support Path objects in the near future
            sftp.put(localpath=str(filepath), remotepath=remote_path)



from pathlib import Path
from typing import Any, Dict, Hashable, Optional, Union, cast

import numpy as np
import xarray as xr


def parse_noaa(
    data_filepath: Union[str, Path],
    site: str,
    measurement_type: str,
    inlet: Optional[str] = None,
    network: str = "NOAA",
    instrument: Optional[str] = None,
    sampling_period: Optional[str] = None,
) -> Dict:
    """Read NOAA data from raw text file or ObsPack NetCDF

    Args:
        data_filepath: Data filepath
        site: Three letter site code
        inlet: Inlet height, if no height use measurement type e.g. flask
        measurement_type: One of ("flask", "insitu", "pfp")
        network: Network, defaults to NOAA
        instrument: Instrument name
        sampling_period: Sampling period
    Returns:
        dict: Dictionary of data and metadata
    """
    if sampling_period is None:
        sampling_period = "NOT_SET"

    sampling_period = str(sampling_period)

    file_extension = Path(data_filepath).suffix

    if file_extension == ".nc":
        return _read_obspack(
            data_filepath=data_filepath,
            site=site,
            inlet=inlet,
            measurement_type=measurement_type,
            instrument=instrument,
            sampling_period=sampling_period,
        )
    else:
        return _read_raw_file(
            data_filepath=data_filepath,
            site=site,
            inlet=inlet,
            measurement_type=measurement_type,
            instrument=instrument,
            sampling_period=sampling_period,
        )


def _format_inlet(inlet: Union[str, float, int]) -> str:
    """
    Output inlet name in expected format. This will include 1 decimal place for floats with a fractional component
    and 0 decimal places otherwise.

    Args:
        inlet : Inlet value. This can be a string as long as this can be converted directly to a float

    Returns:
        str : Formatted string of inlet name

    Examples:
        >>> _format_inlet(40)
        "40m"
        >>> _format_inlet("10.0")
        "10m"
        >>> _format_inlet(2.511432)
        "2.5m"
    """
    if isinstance(inlet, str):
        try:
            inlet_value: Union[float, np.floating] = float(inlet)
        except ValueError:
            inlet_values_str = inlet.split("-")
            try:
                inlet_values = [float(value) for value in inlet_values_str]
            except ValueError:
                raise ValueError("Unable to convert inlet {inlet} to a float")
            else:
                inlet_value = np.mean(inlet_values)
    else:
        inlet_value = inlet

    if inlet_value % 1 == 0:
        inlet_str = f"{inlet_value:.0f}m"
    else:
        inlet_str = f"{inlet_value:.1f}m"

    return inlet_str


def _standarise_variables(obspack_ds: xr.Dataset, species: str) -> xr.Dataset:
    """
    Converts data from NOAA ObsPack dataset into our standardised variables to be stored within the object store.
    The species is also needed so this name can be used to label the variables in the new dataset.

    Expects inputs with: "value", "value_std_dev" or "value_unc", "nvalue" as per NOAA ObsPack standard.

    Args:
        obspack_ds : Dataset derived from a netcdf file within the NOAA obspack
        species : Standardised species name (e.g. "ch4")

    Returns:
        Dataset : Standardised dataset with variables extracted and renamed

    Example output:
        For species = "ch4":
            xarray.Dataset("ch4":[...]
                           "ch4_variability":[...]
                           "ch4_number_of_observations": [...])
    """

    processed_ds = obspack_ds.copy()

    # Rename variables to match our internal standard
    # "value_std_dev" --> f"{species}_variability"
    # "value_unc" --> ??
    # TODO: Clarify what "value_unc" should be renamed to

    variable_names = {
        "value": species,
        "value_std_dev": f"{species}_variability",
        "value_unc": f"{species}_variability",  # May need to be updated
        "nvalue": f"{species}_number_of_observations",
    }

    to_extract = [name for name in variable_names.keys() if name in obspack_ds]

    # For the error variables we only want to take one set of values from the
    # obspack dataset but multiple variables may be available.
    # If multiple are present, combine these together and only extract one
    error_names = ["value_std_dev", "value_unc"]
    error_variables = [name for name in error_names if name in to_extract]

    if len(error_variables) > 1:
        main_ev = error_variables[0]  # Treat first item in the list at the one to keep

        history_attr = "history"
        processed_ds[main_ev].attrs[history_attr] = f"Merged {main_ev} variable from original file with "

        for ev in error_variables[1:]:
            # Combine details from additional additional error variable with main variable
            variable = processed_ds[main_ev]
            new_variable = processed_ds[ev]

            # Update Dataset and add details within attributes
            updated_variable = variable.combine_first(new_variable)
            processed_ds[main_ev] = updated_variable
            processed_ds[main_ev].attrs[history_attr] += f"{ev}, "

            # Remove this extra variables from the list of variables to extract from the dataset
            to_extract.remove(ev)

    # Create dictionary of names to convert obspack ds to our format
    name_dict = {name: key for name, key in variable_names.items() if name in to_extract}

    if not to_extract:
        wanted = variable_names.keys()
        raise ValueError(
            f"No valid data variables columns found in obspack dataset. We expect the following data variables in the passed NetCDF: {wanted}"
        )

    # Grab only the variables we want to keep and rename these
    processed_ds = processed_ds[to_extract]
    processed_ds = processed_ds.rename(name_dict)
    processed_ds = processed_ds.sortby("time")

    return processed_ds


def _split_inlets(
    obspack_ds: xr.Dataset, attributes: Dict, metadata: Dict, inlet: Optional[str] = None
) -> Dict:
    """
    Splits the overall dataset by different inlet values, if present. The expected dataset input should be from the NOAA ObsPack.

    Args:
        obspack_ds : Dataset derived from a netcdf file within the NOAA obspack
        attributes: Attributes extracted from the NOAA obspack. Should contain at least "species" and "measurement_type"
        metadata: Metadata to store alongside standardised data

    Returns:
        Dict: gas data containing "data", "metadata", "attributes" for each inlet

    Example output:
        {"ch4": {"data": xr.Dataset(...), "attributes": {...}, "metadata": {...}}}
        or
        {"ch4_40m": {"data": xr.Dataset(...), "attributes": {...}, "metadata": {...}}, "ch4_60m": {...}, ...}

    """

    orig_attrs = obspack_ds.attrs
    species = attributes["species"]
    measurement_type = attributes["measurement_type"]

    height_var = "intake_height"

    # Check whether the input data contains different inlet height values for each data point ("intake_height" data variable)
    # If so we need to select the data for each inlet and indicate this is a separate Datasource
    # Each data key is labelled based on the species and the inlet (if needed)

    gas_data: Dict[str, Dict] = {}
    if height_var in obspack_ds.data_vars:

        if inlet is not None:
            # TODO: Add to logging?
            print(
                f"WARNING: Ignoring inlet value of {inlet} since file has each data point has an associated height (contains 'intake_height' variable)"
            )

        # Group dataset by the height values
        # Note: could use ds.groupby_bins(...) if necessary if there are lots of small height differences to group these
        obspack_ds_grouped = obspack_ds.groupby(height_var)
        num_groups = len(obspack_ds_grouped.groups)

        # For each group standardise and store with id based on species and inlet height
        for ht, obspack_ds_ht in obspack_ds_grouped:

            # Creating id keys of the form "<species>_<inlet>" e.g. "ch4_40m" or "co_12.5m"
            inlet_str = _format_inlet(ht)
            inlet_num_str = inlet_str.strip("m")

            if num_groups > 1:
                id_key = f"{species}_{inlet_str}"
            else:
                id_key = f"{species}"

            # Extract wanted variables and convert to standardised names
            standarised_ds = _standarise_variables(obspack_ds_ht, species)

            gas_data[id_key] = {}
            gas_data[id_key]["data"] = standarised_ds

            # Add inlet details to attributes and metadata
            attrs_copy = attributes.copy()
            meta_copy = metadata.copy()

            attrs_copy["inlet"] = inlet_str
            attrs_copy["inlet_height_magl"] = inlet_num_str
            meta_copy["inlet"] = inlet_str
            meta_copy["inlet_height_magl"] = inlet_num_str

            gas_data[id_key]["metadata"] = meta_copy
            gas_data[id_key]["attributes"] = attrs_copy

    else:
        try:
            inlet_value = orig_attrs["dataset_intake_ht"]
        except KeyError:
            inlet_from_file = None
        else:
            inlet_from_file = _format_inlet(inlet_value)

        if measurement_type == "flask":
            inlet_from_file = "flask"

        # Check inlet from file against any provided inlet
        if inlet is None and inlet_from_file:
            inlet = inlet_from_file
        elif inlet is not None and inlet_from_file:
            if inlet != inlet_from_file:
                print(
                    f"WARNING: Provided inlet {inlet} does not match inlet derived from the input file: {inlet_from_file}"
                )
        else:
            raise ValueError(
                "Unable to derive inlet from NOAA file. Please pass as an input. If flask data pass 'flask' as inlet."
            )

        id_key = f"{species}"

        if inlet != "flask":
            inlet_num_str = inlet.strip("m")
        else:
            inlet_num_str = ""

        metadata["inlet"] = inlet
        metadata["inlet_height_magl"] = inlet_num_str
        attributes["inlet"] = inlet
        attributes["inlet_height_magl"] = inlet_num_str

        standardised_ds = _standarise_variables(obspack_ds, species)

        gas_data[id_key] = {"data": standardised_ds, "metadata": metadata, "attributes": attributes}

    return gas_data


def _read_obspack(
    data_filepath: Union[str, Path],
    site: str,
    sampling_period: str,
    measurement_type: str,
    inlet: Optional[str] = None,
    instrument: Optional[str] = None,
) -> Dict[str, Dict]:
    """Read NOAA ObsPack NetCDF files

    Args:
        data_filepath: Path to file
        site: Three letter site code
        sampling_period: Sampling period
        measurement_type: One of flask, insitu or pfp
        inlet: Inlet height, if no height use measurement type e.g. flask
        instrument: Instrument name

    Returns:
        dict: Dictionary of results
    """
    from openghg.standardise.meta import assign_attributes
    from openghg.util import clean_string

    valid_types = ("flask", "insitu", "pfp")

    if measurement_type not in valid_types:
        raise ValueError(f"measurement_type must be one of {valid_types}")

    with xr.open_dataset(data_filepath) as temp:
        obspack_ds = temp
        orig_attrs = temp.attrs

    # Want to find and drop any duplicate time values for the original dataset
    # Using xarray directly we have to do in a slightly convoluted way as this is not well built
    # into the xarray workflow yet - https://github.com/pydata/xarray/pull/5239
    # - can use da.drop_duplicates() but only on one variable at a time and not on the whole Dataset
    # This method keeps attributes for each of the variables including units

    # The dimension within the original dataset is called "obs" and has no associated coordinates
    # Extract time from original Dataset (dimension is "obs")
    time = obspack_ds.time

    # To keep associated "obs" dimension, need to assign coordinate values to this (just 0, len(obs))
    time = time.assign_coords(obs=obspack_ds.obs)

    # Make "time" the primary dimension (while retaining "obs") and add "time" values as coordinates
    time = time.swap_dims(dims_dict={"obs": "time"})
    time = time.assign_coords(time=time)

    # Drop any duplicate time values and extract the associated "obs" values
    # TODO: Work out what to do with duplicates - may be genuine multiple measurements
    time_unique = time.drop_duplicates(dim="time", keep="first")
    obs_unique = time_unique.obs

    # Use these obs values to filter the original dataset to remove any repeated times
    processed_ds = obspack_ds.sel(obs=obs_unique)
    processed_ds = processed_ds.set_coords(["time"])

    # Estimate sampling period using metadata and midpoint time
    if sampling_period == "NOT_SET":
        sampling_period_estimate = _estimate_sampling_period(obspack_ds)
    else:
        sampling_period_estimate = -1.0

    species = clean_string(obspack_ds.attrs["dataset_parameter"])
    network = "NOAA"

    try:
        # Extract units attribute from value data variable
        units = processed_ds["value"].units
    except (KeyError, AttributeError):
        print("Unable to extract units from 'value' within input dataset")

    metadata = {}
    metadata["site"] = site
    metadata["network"] = network
    metadata["measurement_type"] = measurement_type
    metadata["species"] = species
    metadata["units"] = units
    metadata["sampling_period"] = sampling_period
    metadata["data_source"] = "noaa_obspack"

    # Add additional sampling_period_estimate if sampling_period is not set
    if sampling_period_estimate >= 0.0:
        metadata["sampling_period_estimate"] = str(
            sampling_period_estimate
        )  # convert to string to keep consistent with "sampling_period"

    # Add instrument if present
    if instrument is not None:
        metadata["instrument"] = instrument
    else:
        metadata["instrument"] = orig_attrs.get("instrument", "NOT_SET")

    # Add data owner details, station position and calibration scale, if present
    metadata["data_owner"] = orig_attrs.get("provider_1_name", "NOT_SET")
    metadata["data_owner_email"] = orig_attrs.get("provider_1_email", "NOT_SET")
    metadata["station_longitude"] = orig_attrs.get("site_longitude", "NOT_SET")
    metadata["station_latitude"] = orig_attrs.get("site_latitude", "NOT_SET")
    metadata["calibration_scale"] = orig_attrs.get("dataset_calibration_scale", "NOT_SET")

    # Create attributes with copy of metadata values
    attributes = cast(Dict[Hashable, Any], metadata.copy())  # Cast to match xarray attributes type

    # TODO: At the moment all attributes from the NOAA ObsPack are being copied
    # plus any variables we're adding - decide if we want to reduce this
    attributes.update(orig_attrs)

    # expected_keys = {
    #     "site",
    #     "species",
    #     "inlet",
    #     "instrument",
    #     "sampling_period",
    #     "calibration_scale",
    #     "station_longitude",
    #     "station_latitude",
    # }

    gas_data = _split_inlets(processed_ds, attributes, metadata, inlet=inlet)

    gas_data = assign_attributes(data=gas_data, site=site, network=network)

    return gas_data


def _read_raw_file(
    data_filepath: Union[str, Path],
    site: str,
    sampling_period: str,
    measurement_type: str,
    inlet: Optional[str] = None,
    instrument: Optional[str] = None,
) -> Dict:
    """Reads NOAA data files and returns a dictionary of processed
    data and metadata.

    Args:
        data_filepath: Path of file to load
        site: Site name
        sampling_period: Sampling period
        measurement_type: One of flask, insitu or pfp
        inlet: Inlet height, if no height use measurement type e.g. flask
        instrument: Instrument name

    Returns:
        list: UUIDs of Datasources data has been assigned to
    """
    from openghg.standardise.meta import assign_attributes

    # TODO: Added this for now to make sure inlet is specified but may be able to remove
    # if this can be derived from the data format.
    if inlet is None:
        raise ValueError("Inlet must be specified to derive data from NOAA raw (txt) files.")

    data_filepath = Path(data_filepath)
    filename = data_filepath.name

    species = filename.split("_")[0].lower()

    source_name = data_filepath.stem
    source_name = source_name.split("-")[0]

    gas_data = _read_raw_data(
        data_filepath=data_filepath,
        inlet=inlet,
        species=species,
        measurement_type=measurement_type,
        sampling_period=sampling_period,
    )

    gas_data = assign_attributes(data=gas_data, site=site, network="NOAA")

    return gas_data


def _read_raw_data(
    data_filepath: Path,
    species: str,
    inlet: str,
    sampling_period: str,
    measurement_type: str = "flask",
) -> Dict:
    """Separates the gases stored in the dataframe in
    separate dataframes and returns a dictionary of gases
    with an assigned UUID as gas:UUID and a list of the processed
    dataframes

    Args:
        data_filepath: Path of datafile
        species: Species string such as CH4, CO
        measurement_type: Type of measurements e.g. flask
    Returns:
        dict: Dictionary containing attributes, data and metadata keys
    """
    from openghg.util import clean_string, load_json, read_header
    from pandas import Timestamp, read_csv

    header = read_header(filepath=data_filepath)

    column_names = header[-1][14:].split()

    def date_parser(year: str, month: str, day: str, hour: str, minute: str, second: str) -> Timestamp:
        return Timestamp(year, month, day, hour, minute, second)

    date_parsing = {
        "time": [
            "sample_year",
            "sample_month",
            "sample_day",
            "sample_hour",
            "sample_minute",
            "sample_seconds",
        ]
    }

    data_types = {
        "sample_year": int,
        "sample_month": int,
        "sample_day": int,
        "sample_hour": int,
        "sample_minute": int,
        "sample_seconds": int,
    }

    # Number of header lines to skip
    n_skip = len(header)

    data = read_csv(
        data_filepath,
        skiprows=n_skip,
        names=column_names,
        sep=r"\s+",
        dtype=data_types,
        parse_dates=date_parsing,
        date_parser=date_parser,
        index_col="time",
        skipinitialspace=True,
    )

    # Drop duplicates
    data = data.loc[~data.index.duplicated(keep="first")]

    # Check if the index is sorted
    if not data.index.is_monotonic_increasing:
        data = data.sort_index()

    # Read the site code from the Dataframe
    site = str(data["sample_site_code"][0]).upper()

    site_data = load_json("acrg_site_info.json")
    # If this isn't a site we recognize try and read it from the filename
    if site not in site_data:
        site = str(data_filepath.name).split("_")[1].upper()

        if site not in site_data:
            raise ValueError(f"The site {site} is not recognized.")

    if species is not None:
        # If we're passed a species ensure that it is in fact the correct species
        data_species = str(data["parameter_formula"].values[0]).lower()

        passed_species = species.lower()
        if data_species != passed_species:
            raise ValueError(
                f"Mismatch between passed species ({passed_species}) and species read from data ({data_species})"
            )

    species = species.upper()

    flag = []
    selection_flag = []
    for flag_str in data.analysis_flag:
        flag.append(flag_str[0] == ".")
        selection_flag.append(int(flag_str[1] != "."))

    combined_data = {}

    data[species + "_status_flag"] = flag
    data[species + "_selection_flag"] = selection_flag

    data = data[data[species + "_status_flag"]]

    data = data[
        [
            "sample_latitude",
            "sample_longitude",
            "sample_altitude",
            "analysis_value",
            "analysis_uncertainty",
            species + "_selection_flag",
        ]
    ]

    rename_dict = {
        "analysis_value": species,
        "analysis_uncertainty": species + "_repeatability",
        "sample_longitude": "longitude",
        "sample_latitude": "latitude",
        "sample_altitude": "altitude",
    }

    data = data.rename(columns=rename_dict, inplace=False)
    data = data.to_xarray()

    # TODO  - this could do with a better name
    noaa_params = load_json("attributes.json")["NOAA"]

    site_attributes = noaa_params["global_attributes"]
    site_attributes["inlet_height_magl"] = "NA"
    site_attributes["instrument"] = noaa_params["instrument"][species.upper()]
    site_attributes["sampling_period"] = sampling_period

    metadata = {}
    metadata["species"] = clean_string(species)
    metadata["site"] = site
    metadata["measurement_type"] = measurement_type
    metadata["network"] = "NOAA"
    metadata["inlet"] = inlet
    metadata["sampling_period"] = sampling_period
    metadata["instrument"] = noaa_params["instrument"][species.upper()]

    combined_data[species.lower()] = {
        "metadata": metadata,
        "data": data,
        "attributes": site_attributes,
    }

    return combined_data


def _estimate_sampling_period(obspack_ds: xr.Dataset, min_estimate: float = 10.0) -> float:
    """
    Estimate the sampling period for the NOAA data using either the "data_selection_tag"
    attribute (this sometimes contains useful information such as "HourlyData") or by using
    the midpoint_time within the data itself.

    Note: midpoint_time often seems to match start_time implying instantaneous measurement
    or that this value has not been correctly included.

    If the estimate is less than `min_estimate` the estimate sampling period will be set to
    this value.

    Args:
        obspack_ds : Dataset of raw obs pack file opened as an xarray Dataset
        min_estimate : Minimum sampling period estimate to use in seconds.

    Returns:
        int: Seconds for the estimated sampling period.
    """
    # Check useful attributes
    data_selection = obspack_ds.attrs["dataset_selection_tag"]

    hourly_s = 60 * 60
    daily_s = hourly_s * 24
    weekly_s = daily_s * 7
    monthly_s = weekly_s * 28  # approx
    yearly_s = daily_s * 365  # approx

    sampling_period_estimate = 0.0  # seconds

    frequency_keywords = {
        "hourly": hourly_s,
        "daily": daily_s,
        "weekly": weekly_s,
        "monthly": monthly_s,
        "yearly": yearly_s,
    }
    for freq, time_s in frequency_keywords.items():
        if freq in data_selection.lower():
            sampling_period_estimate = time_s

    if not sampling_period_estimate:
        if "start_time" in obspack_ds and "midpoint_time" in obspack_ds:
            half_sample_time = (obspack_ds["midpoint_time"] - obspack_ds["start_time"]).values
            half_sample_time_s = half_sample_time.astype("timedelta64[s]").mean().astype(float)
            sampling_period_estimate = round(half_sample_time_s * 2, 1)

    if sampling_period_estimate < min_estimate:
        sampling_period_estimate = min_estimate

    return sampling_period_estimate



from pathlib import Path
from typing import Dict, Optional, Union

pathType = Union[str, Path]


def parse_aqmesh(
    data_filepath: pathType,
    metadata_filepath: pathType,
    sampling_period: Optional[str] = None,
    **kwargs: Dict,
) -> Dict:
    """Read AQMesh data files

    Args:
        data_filepath: Data filepath
        metadata_filepath: Metadata filepath
        sampling_period: Measurement sampling period (str)
    Returns:
        dict: Dictionary of data
    """
    from addict import Dict as aDict
    from pandas import read_csv

    if sampling_period is None:
        sampling_period = "NOT_SET"

    use_cols = [0, 1, 4, 6]
    datetime_cols = {"time": ["date_UTC"]}
    na_values = [-999, -999.0]

    df = read_csv(
        data_filepath,
        index_col="time",
        usecols=use_cols,
        parse_dates=datetime_cols,
        na_values=na_values,
    )

    # This might change so we'll read it each time for now
    metadata = _parse_metadata(filepath=metadata_filepath)

    # Species is given in the data column
    orig_species = df.columns[0]
    species_split = orig_species.split("_")

    species = species_split[0]
    units = species_split[1]

    species_lower = species.lower()
    rename_cols = {orig_species: species_lower, "location_name": "site"}
    df = df.rename(columns=rename_cols)
    df = df.dropna(axis="rows", subset=[species_lower])

    # TODO - add in assignment of attributes
    # assign_attributes

    site_groups = df.groupby(df["site"])
    site_data = aDict()
    for site, site_df in site_groups:
        site_name = site.replace(" ", "").lower()
        site_df = site_df.drop("site", axis="columns")
        site_data[site_name]["data"] = site_df.to_xarray()
        site_data[site_name]["metadata"] = metadata[site_name]
        # Add in the species to the metadata
        site_data[site_name]["metadata"]["species"] = species_lower
        site_data[site_name]["metadata"]["units"] = units
        site_data[site_name]["metadata"]["sampling_period"] = sampling_period

    site_dict: Dict = site_data.to_dict()
    return site_dict


def _parse_metadata(filepath: pathType) -> Dict:
    """Parse AQMesh metadata

    Args:
        filepath: Path to metadata CSV
        pipeline: If running in pipeline skip the writing of metadata to file
    Returns:
        dict: Dictionary of metadata
    """
    from addict import Dict as aDict
    from openghg.util import check_date
    from pandas import read_csv

    filepath = Path(filepath)
    raw_metadata = read_csv(filepath)

    site_metadata = aDict()

    for _, row in raw_metadata.iterrows():
        site_name = row["location_name"].replace(" ", "").lower()
        site_data = site_metadata[site_name]

        site_data["site"] = site_name
        site_data["pod_id"] = row["pod_id_location"]
        site_data["start_date"] = check_date(row["start_date_UTC"])
        site_data["end_date"] = check_date(row["end_date_UTC"])
        site_data["relocate_date"] = check_date(row["relocate_date_UTC"])
        site_data["long_name"] = row["location_name"]
        site_data["borough"] = row["Borough"]
        site_data["site_type"] = row["Type"]
        site_data["in_ulez"] = row["ULEZ"]
        site_data["latitude"] = row["Latitude"]
        site_data["longitude"] = row["Longitude"]
        site_data["inlet"] = row["Height"]
        site_data["network"] = "aqmesh_glasgow"
        site_data["sampling_period"] = "NA"

    # TODO - I feel this is a bit clunky
    dict_metadata: Dict = site_metadata.to_dict()
    return dict_metadata



from typing import Dict, Union

from openghg.store import ObsSurface
from openghg.store.base import Datasource
from openghg.util import create_daterange_str, verify_site


class RankSources:
    def __init__(self, cloud: bool = False, service_url: str = None) -> None:
        self._cloud = cloud

        if cloud:
            raise NotImplementedError
            # from Acquire.Client import Wallet

            # wallet = Wallet()

            # if service_url is None:
            #     service_url = "https://fn.openghg.org/t"

            # self._service = wallet.get_service(service_url=f"{service_url}/openghg")

    def raw(self) -> Dict:
        """Return the raw ranking data

        Args:
            None
        Returns:
            dict: Raw ranking data
        """
        return self._user_info

    def get_sources(self, site: str, species: str) -> Dict:
        """Get the datasources for this site and species to allow a ranking to be set

        Args:
            site: Three letter site code
            species: Species name
        Returns:
            dict: Dictionary of datasource metadata
        """
        if self._cloud:
            raise NotImplementedError
            # return self._get_sources_cloud(site=site, species=species)
        else:
            return self._get_sources_local(site=site, species=species)

    # def _get_sources_cloud(self, site: str, species: str) -> Dict:
    #     site = verify_site(site=site)

    #     args = {"site": site, "species": species}

    #     self.site = site
    #     self.species = species

    #     response: Dict = self._service.call_function(function="rank.get_sources", args=args)

    #     if not response:
    #         raise ValueError(f"No sources found for {species} at {site}")

    #     self._user_info: Dict = response["user_info"]
    #     self._key_lookup: Dict = response["key_lookup"]
    #     self._needs_update = False

    #     return self._user_info

    def _get_sources_local(self, site: str, species: str) -> Dict:
        site = verify_site(site=site)

        # Save these
        self.site = site
        self.species = species

        obs = ObsSurface.load()
        datasource_uuids = obs.datasources()
        rank_table = obs.rank_data()

        # Shallow load the Datasources (only get their JSON metadata)
        datasources = (Datasource.load(uuid=uuid, shallow=True) for uuid in datasource_uuids)

        matching_sources = [d for d in datasources if d.search_metadata(site=site, species=species)]

        if matching_sources:
            self._user_info = {
                d.inlet(): {
                    "rank_data": rank_table.get(d.uuid(), "NA"),
                    "data_range": d.daterange_str(),
                }
                for d in matching_sources
            }

            self._key_lookup = {d.inlet(): d.uuid() for d in matching_sources}
            self._needs_update = False
        else:
            self._user_info = {}

        return self._user_info

    def get_specific_source(self, inlet: str) -> Dict:
        """Return the ranking data of a specific key

        Args:
            key: Key
        Returns:
            dict: Dictionary of ranking data
        """
        if self._needs_update:
            self.get_sources(site=self.site, species=self.species)

        rank_data: Dict[str, Union[str, Dict]] = self._user_info[inlet]["rank_data"]
        return rank_data

    def set_rank(
        self,
        inlet: str,
        rank: Union[int, str],
        start_date: str,
        end_date: str,
        overwrite: bool = False,
    ) -> None:
        """Set the rank data for the

        Args:
            inlet: Inlet to set ranking data
            rank: Number between 1 and 9
            start_date: Start date
            end_date: End date
            overwrite: If True overwrite current ranking data
        Returns:
            None
        """
        if self._cloud:
            raise NotImplementedError
        else:
            return self._set_rank_local(
                inlet=inlet,
                rank=rank,
                start_date=start_date,
                end_date=end_date,
                overwrite=overwrite,
            )

    # def _set_rank_cloud(
    #     self,
    #     inlet: str,
    #     rank: Union[int, str],
    #     start_date: str,
    #     end_date: str,
    #     overwrite: bool = False,
    # ) -> None:
    #     inlet = inlet.lower()
    #     uuid = self._key_lookup[inlet]

    #     dateranges = create_daterange_str(start=start_date, end=end_date)

    #     args: Dict[str, Union[str, int, List]] = {}
    #     args["rank"] = rank
    #     args["uuid"] = uuid
    #     args["dateranges"] = dateranges
    #     args["overwrite"] = overwrite

    #     self._service.call_function(function="rank.set_rank", args=args)
    #     self._needs_update = True

    def _set_rank_local(
        self,
        inlet: str,
        rank: Union[int, str],
        start_date: str,
        end_date: str,
        overwrite: bool = False,
    ) -> None:
        obs = ObsSurface.load()

        inlet = inlet.lower()
        uuid = self._key_lookup[inlet]

        daterange = create_daterange_str(start=start_date, end=end_date)

        obs.set_rank(uuid=uuid, rank=rank, date_range=daterange, overwrite=overwrite)

        self._needs_update = True

    def clear_rank(self, inlet: str) -> None:
        """Clear the ranking data for a Datasource

        Args:
            key: Key for specific source
        Returns:
            None
        """
        if self._cloud:
            raise NotImplementedError
            # return self._clear_rank_cloud(inlet=inlet)
        else:
            return self._clear_rank_local(inlet=inlet)

    def _clear_rank_local(self, inlet: str) -> None:
        """Clear the ranking data for a Datasource

        Args:
            key: Key for specific source
        Returns:
            None
        """
        obs = ObsSurface.load()
        inlet = inlet.lower()
        uuid = self._key_lookup[inlet]
        obs.clear_rank(uuid=uuid)
        self._needs_update = True

    # def _clear_rank_cloud(self, inlet: str) -> None:
    #     """Clear the ranking data for a Datasource

    #     Args:
    #         key: Key for specific source
    #     Returns:
    #         None
    #     """
    #     uuid = self._key_lookup[inlet]
    #     args = {"uuid": uuid}
    #     self._service.call_function(function="rank.clear_rank", args=args)
    #     self._needs_update = True

    # def visualise_rankings(self) -> Network:
    #     """ Creates a small network graph of ranked data with each rank given a colour

    #         Note that this function should only be run from a Jupyter Notebook

    #         Args:
    #             rank_data (dict): Dictionary of the form given by RankSources.get_sources()
    #         Returns:
    #             pyvis.network.Network: Network graph
    #     """
    #     header_text = "OpenGHG ranked data"
    #     net = Network("800px", "100%", notebook=True, heading=header_text)
    #     # Set the physics layout of the network
    #     net.force_atlas_2based()

    #     rank_data = self._key_lookup

    #     a_key = list(rank_data.keys())[0]
    #     site = rank_data[a_key]["metadata"]["site"].upper()

    #     norm = matplotlib.colors.Normalize(vmin=0, vmax=10, clip=True)
    #     mapper = cm.ScalarMappable(norm=norm, cmap=cm.tab10)

    #     def colour_mapper(x):
    #         return matplotlib.colors.to_hex(mapper.to_rgba(int(x)))

    #     net.add_node(site, label=site, color="brown", value=5000)

    #     for key, data in rank_data.items():
    #         rank = data["rank"]
    #         site_name = data["metadata"]["site"].upper()
    #         data_range = data["data_range"]

    #         # HTML to show when the mouse is hovered over a node
    #         title = "</br>".join([f"<b>Rank:</b> {str(rank)}", f"<b>Site:</b> {site_name}", f"<b>Data range:</b> {data_range}"])

    #         if rank == 0:
    #             colour = colour_mapper(rank)
    #         else:
    #             # For now just use the highest rank for the color
    #             highest_rank = sorted(list(rank.keys()))[-1]
    #             colour = colour_mapper(highest_rank)

    #         split_key = key.split("_")
    #         label = " ".join((split_key[0].upper(), split_key[1].upper(), split_key[2], split_key[3]))

    #         net.add_node(key, label=label, title=title, color=colour, value=2000)
    #         net.add_edge(source=site, to=key)

    #     return net.show("openghg_rankings.html")



from xarray import Dataset

__all__ = ["plot_footprint"]


def plot_footprint(data: Dataset, label: str = None, vmin: float = None, vmax: float = None) -> None:
    """Plot a footprint

    Args:
        data: Dataset containing fp variable
        label: Label for colourbar
        vmin: Minimum value for colours
        vmax: MinimumMax value for colours
    Returns:
        None
    """
    import matplotlib.colors as colors
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()

    # Plot footprints as a 2D colour map
    data_fp = data.fp.isel(time=0)  # First time point
    lat = data_fp.lat
    lon = data_fp.lon
    footprint = data_fp.values

    vmin = 1e-5  # min is 0 and can't use 0 for a log scale
    vmax = footprint.max()

    im = ax.pcolormesh(
        lon, lat, footprint, norm=colors.LogNorm(vmin=vmin, vmax=vmax), shading="auto"
    )  # Put on a log scale
    cb = plt.colorbar(im, ax=ax)

    if label:
        cb.set_label(label)



from __future__ import annotations

# import cdsapi  # type: ignore
# import numpy as np
from typing import TYPE_CHECKING, List, Optional, Union

# import requests
# from xarray import open_dataset, Dataset
# from requests.adapters import HTTPAdapter
# from requests.packages.urllib3.util.retry import Retry
# from openghg.util import timestamp_tzaware

if TYPE_CHECKING:
    from openghg.dataobjects import METData

__all__ = ["retrieve_met", "METData"]


def retrieve_met(
    site: str,
    network: str,
    years: Union[str, List[str]],
    variables: Optional[List[str]] = None,
) -> METData:
    """Retrieve METData data. Note that this function will only download a
    full year of data which may take some time.

    This function currently on retrieves data from the "reanalysis-era5-pressure-levels"
    dataset but may be modified for other datasets in the future.
    Args:
        site: Three letter sitec code
        network: Network
        years: Year(s) of data required
    Returns:
        METData: METData object holding data and metadata
    """
    raise NotImplementedError("The met retrieval module needs updating and doesn't currently work.")
    # from openghg.dataobjects import METData

    # if variables is None:
    #     variables = ["u_component_of_wind", "v_component_of_wind"]

    # latitude, longitude, site_height, inlet_heights = _get_site_data(site=site, network=network)

    # # Get the area to retrieve data for
    # ecmwf_area = _get_ecmwf_area(site_lat=latitude, site_long=longitude)
    # # Calculate the pressure at measurement height(s)
    # measure_pressure = _get_site_pressure(inlet_heights=inlet_heights, site_height=site_height)
    # # Calculate the ERA5 pressure levels required
    # ecmwf_pressure_levels = _altitude_to_ecmwf_pressure(measure_pressure=measure_pressure)

    # if not isinstance(years, list):
    #     years = [years]
    # else:
    #     years = sorted(years)

    # # TODO - we might need to customise this further in the future to
    # # request other types of weather data
    # request = {
    #     "product_type": "reanalysis",
    #     "format": "netcdf",
    #     "variable": variables,
    #     "pressure_level": ecmwf_pressure_levels,
    #     "year": [str(x) for x in years],
    #     "month": [str(x).zfill(2) for x in range(1, 13)],
    #     "day": [str(x).zfill(2) for x in range(1, 32)],
    #     "time": [f"{str(x).zfill(2)}:00" for x in range(0, 24)],
    #     "area": ecmwf_area,
    # }

    # cds_client = cdsapi.Client()
    # dataset_name = "reanalysis-era5-pressure-levels"

    # # Retrieve metadata from Copernicus about the dataset, this includes
    # # the location of the data netCDF file.
    # result = cds_client.retrieve(name=dataset_name, request=request)

    # # Download the data itself
    # dataset = _download_data(url=result.location)
    # # dataset = xr.open_dataset("/home/gar/Documents/Devel/RSE/openghg/tests/data/request_return.nc")

    # # We replace the date data with a start and end date here
    # start_date = str(timestamp_tzaware(f"{years[0]}-1-1"))
    # end_date = str(timestamp_tzaware(f"{years[-1]}-12-31"))

    # metadata = {
    #     "product_type": request["product_type"],
    #     "format": request["format"],
    #     "variable": request["variable"],
    #     "pressure_level": request["pressure_level"],
    #     "area": request["area"],
    #     "site": site,
    #     "network": network,
    #     "start_date": start_date,
    #     "end_date": end_date,
    # }

    # return METData(data=dataset, metadata=metadata)


# def _download_data(url: str) -> Dataset:
#     """Retrieve data from the passed URL. This is used to retrieve data from
#     the Copernicus data store.

#     Args:
#         url: URL string
#     Returns:
#         xarray.Dataset: NetCDF data retrieved
#     """
#     # If we get any of these codes we'll try again
#     retriable_status_codes = [
#         requests.codes.internal_server_error,
#         requests.codes.bad_gateway,
#         requests.codes.service_unavailable,
#         requests.codes.gateway_timeout,
#         requests.codes.too_many_requests,
#         requests.codes.request_timeout,
#     ]

#     timeout = 20  # seconds

#     retry_strategy = Retry(
#         total=3,
#         status_forcelist=retriable_status_codes,
#         allowed_methods=["HEAD", "GET", "OPTIONS"],
#         backoff_factor=1,
#     )  # type: ignore

#     adapter = HTTPAdapter(max_retries=retry_strategy)

#     http = requests.Session()
#     http.mount("https://", adapter)
#     http.mount("http://", adapter)

#     data = http.get(url, timeout=timeout).content

#     try:
#         dataset: Dataset = open_dataset(data)
#     except ValueError:
#         raise ValueError("Invalid data returned, cannot create dataset.")

#     return dataset


# def _two_closest_values(diff: np.ndarray) -> np.ndarray:
#     """Get location of two closest values in an array of differences.

#     Args:
#         diff: Numpy array of values
#     Returns:
#         np.ndarry: Numpy array of two closes values
#     """
#     closest_values: np.ndarray = np.argpartition(np.abs(diff), 2)[:2]
#     return closest_values


# def _get_site_data(site: str, network: str) -> Tuple[float, float, float, List]:
#     """Extract site location data from site attributes file.

#     Args:
#         site: Site code
#     Returns:
#         dict: Dictionary of site data
#     """
#     from openghg.util import load_json

#     network = network.upper()
#     site = site.upper()

#     site_info = load_json("acrg_site_info.json")

#     try:
#         site_data = site_info[site][network]
#         latitude = float(site_data["latitude"])
#         longitute = float(site_data["longitude"])
#         site_height = float(site_data["height_station_masl"])
#         inlet_heights = site_data["height_name"]
#     except KeyError as e:
#         raise KeyError(f"Incorrect site or network : {e}")

#     return latitude, longitute, site_height, inlet_heights


# def _get_ecmwf_area(site_lat: float, site_long: float) -> List:
#     """Find out the area required from ERA5.

#     Args:
#         site_lat: Latitude of site
#         site_long: Site longitude
#     Returns:
#         list: List of min/max lat long values
#     """
#     ecwmf_lat = np.arange(-90, 90.25, 0.25)
#     ecwmf_lon = np.arange(-180, 180.25, 0.25)

#     ecwmf_lat_indices = _two_closest_values(ecwmf_lat - site_lat)
#     ecwmf_lon_indices = _two_closest_values(ecwmf_lon - site_long)

#     return [
#         np.max(ecwmf_lat[ecwmf_lat_indices]),
#         np.min(ecwmf_lon[ecwmf_lon_indices]),
#         np.min(ecwmf_lat[ecwmf_lat_indices]),
#         np.max(ecwmf_lon[ecwmf_lon_indices]),
#     ]


# def _get_site_pressure(inlet_heights: List, site_height: float) -> List[float]:
#     """Calculate the pressure levels required

#     Args:
#         inlet_height: Height(s) of inlets
#         site_height: Height of site
#     Returns:
#         list: List of pressures
#     """
#     import re

#     if not isinstance(inlet_heights, list):
#         inlet_heights = [inlet_heights]

#     measured_pressure = []
#     for h in inlet_heights:
#         try:
#             # Extract the number from the inlet height str using regex
#             inlet = float(re.findall(r"\d+(?:\.\d+)?", h)[0])
#             measurement_height = inlet + float(site_height)
#             # Calculate the pressure
#             pressure = 1000 * np.exp((-1 * measurement_height) / 7640)
#             measured_pressure.append(pressure)
#         except IndexError:
#             pass

#     return measured_pressure


# def _altitude_to_ecmwf_pressure(measure_pressure: List[float]) -> List[str]:
#     """Find out what pressure levels are required from ERA5.

#     Args:
#         measure_pressure: List of pressures
#     Returns:
#         list: List of desired pressures
#     """
#     from openghg.util import load_json

#     ecmwf_metadata = load_json("ecmwf_dataset_info.json")
#     dataset_metadata = ecmwf_metadata["datasets"]
#     valid_levels = dataset_metadata["reanalysis_era5_pressure_levels"]["valid_levels"]

#     # Available ERA5 pressure levels
#     era5_pressure_levels = np.array(valid_levels)

#     # Match pressure to ERA5 pressure levels
#     ecwmf_pressure_indices = np.zeros(len(measure_pressure) * 2)

#     for index, m in enumerate(measure_pressure):
#         ecwmf_pressure_indices[(index * 2) : (index * 2 + 2)] = _two_closest_values(m - era5_pressure_levels)

#     desired_era5_pressure = era5_pressure_levels[np.unique(ecwmf_pressure_indices).astype(int)]

#     pressure_levels: List = desired_era5_pressure.astype(str).tolist()

#     return pressure_levels



from collections import defaultdict
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import DefaultDict, Dict, List, NoReturn, Optional, Tuple, Union

import numpy as np
from openghg.store import DataSchema
from openghg.store.base import BaseStore
from pandas import Timestamp
from xarray import Dataset

__all__ = ["Footprints"]


class Footprints(BaseStore):
    """This class is used to process footprints model output"""

    _root = "Footprints"
    _uuid = "62db5bdf-c88d-4e56-97f4-40336d37f18c"
    _metakey = f"{_root}/uuid/{_uuid}/metastore"

    @staticmethod
    def read_data(binary_data: bytes, metadata: Dict, file_metadata: Dict) -> Dict:
        """Ready a footprint from binary data

        Args:
            binary_data: Footprint data
            metadata: Dictionary of metadata
            file_metadat: File metadata
        Returns:
            dict: UUIDs of Datasources data has been assigned to
        """
        with TemporaryDirectory() as tmpdir:
            tmpdir_path = Path(tmpdir)

            try:
                filename = file_metadata["filename"]
            except KeyError:
                raise KeyError("We require a filename key for metadata read.")

            filepath = tmpdir_path.joinpath(filename)
            filepath.write_bytes(binary_data)

            return Footprints.read_file(filepath=filepath, **metadata)

    # @staticmethod
    # def read_data(binary_data: bytes, metadata: Dict, file_metadata: Dict) -> Dict:
    #     """Ready a footprint from binary data

    #     Args:
    #         binary_data: Footprint data
    #         metadata: Dictionary of metadata
    #         file_metadat: File metadata
    #     Returns:
    #         dict: UUIDs of Datasources data has been assigned to
    #     """
    #     from openghg.store import assign_data, infer_date_range, load_metastore, datasource_lookup

    #     fp = Footprints.load()

    #     # Load in the metadata store
    #     metastore = load_metastore(key=fp._metakey)

    #     sha1_hash = file_metadata["sha1_hash"]
    #     overwrite = metadata.get("overwrite", False)

    #     if sha1_hash in fp._file_hashes and not overwrite:
    #         print(
    #             f"This file has been uploaded previously with the filename : {fp._file_hashes[sha1_hash]} - skipping."
    #         )

    #     data_buf = BytesIO(binary_data)
    #     fp_data = open_dataset(data_buf)

    #     fp_time = fp_data.time
    #     period = metadata.get("period")
    #     continuous = metadata["continous"]
    #     high_spatial_res = metadata["high_spatial_res"]
    #     species = metadata["species"]
    #     filename = file_metadata["filename"]

    #     site = metadata["site"]
    #     domain = metadata["domain"]
    #     model = metadata["model"]
    #     height = metadata["height"]

    #     start_date, end_date, period_str = infer_date_range(
    #         fp_time, filepath=filename, period=period, continuous=continuous
    #     )

    #     metadata["start_date"] = str(start_date)
    #     metadata["end_date"] = str(end_date)
    #     metadata["time_period"] = period_str

    #     metadata["max_longitude"] = round(float(fp_data["lon"].max()), 5)
    #     metadata["min_longitude"] = round(float(fp_data["lon"].min()), 5)
    #     metadata["max_latitude"] = round(float(fp_data["lat"].max()), 5)
    #     metadata["min_latitude"] = round(float(fp_data["lat"].min()), 5)

    #     # TODO: Pull out links to underlying data format into a separate format function
    #     #  - high_spatial_res - data vars - "fp_low", "fp_high", coords - "lat_high", "lon_high"
    #     #  - high_time_res - data vars - "fp_HiTRes", coords - "H_back"

    #     metadata["spatial_resolution"] = "standard_spatial_resolution"

    #     if high_spatial_res:
    #         try:
    #             metadata["max_longitude_high"] = round(float(fp_data["lon_high"].max()), 5)
    #             metadata["min_longitude_high"] = round(float(fp_data["lon_high"].min()), 5)
    #             metadata["max_latitude_high"] = round(float(fp_data["lat_high"].max()), 5)
    #             metadata["min_latitude_high"] = round(float(fp_data["lat_high"].min()), 5)

    #             metadata["spatial_resolution"] = "high_spatial_resolution"
    #         except KeyError:
    #             raise KeyError("Expected high spatial resolution. Unable to find lat_high or lon_high data.")

    #     if species == "co2":
    #         # Expect co2 data to have high time resolution
    #         high_time_res = True

    #     metadata["time_resolution"] = "standard_time_resolution"

    #     if high_time_res:
    #         if "fp_HiTRes" in fp_data:
    #             metadata["time_resolution"] = "high_time_resolution"
    #         else:
    #             raise KeyError("Expected high time resolution. Unable to find fp_HiTRes data.")

    #     metadata["heights"] = [float(h) for h in fp_data.height.values]
    #     # Do we also need to save all the variables we have available in this footprints?
    #     metadata["variables"] = list(fp_data.keys())

    #     # if model_params is not None:
    #     #     metadata["model_parameters"] = model_params

    #     # Set the attributes of this Dataset
    #     fp_data.attrs = {"author": "OpenGHG Cloud", "processed": str(timestamp_now())}

    #     # This might seem longwinded now but will help when we want to read
    #     # more than one footprints at a time
    #     key = "_".join((site, domain, model, height))

    #     footprint_data: DefaultDict[str, Dict[str, Union[Dict, Dataset]]] = defaultdict(dict)
    #     footprint_data[key]["data"] = fp_data
    #     footprint_data[key]["metadata"] = metadata

    #     # These are the keys we will take from the metadata to search the
    #     # metadata store for a Datasource, they should provide as much detail as possible
    #     # to uniquely identify a Datasource
    #     required = ("site", "model", "height", "domain")
    #     lookup_results = datasource_lookup(metastore=metastore, data=footprint_data, required_keys=required)

    #     data_type = "footprints"
    #     datasource_uuids: Dict[str, Dict] = assign_data(
    #         data_dict=footprint_data,
    #         lookup_results=lookup_results,
    #         overwrite=overwrite,
    #         data_type=data_type,
    #     )

    #     fp.add_datasources(uuids=datasource_uuids, data=footprint_data, metastore=metastore)

    #     # Record the file hash in case we see this file again
    #     fp._file_hashes[sha1_hash] = filename

    #     fp.save()

    #     metastore.close()

    #     return datasource_uuids

    # def _store_data(data: Dataset, metadata: Dict):
    #     """ Takes an xarray Dataset

    #     Args:
    #         data: xarray Dataset
    #         metadata: Metadata dict
    #     Returns:

    #     """

    @staticmethod
    def read_file(
        filepath: Union[str, Path],
        site: str,
        height: str,
        domain: str,
        model: str,
        metmodel: Optional[str] = None,
        species: Optional[str] = None,
        network: Optional[str] = None,
        period: Optional[Union[str, tuple]] = None,
        continuous: bool = True,
        retrieve_met: bool = False,
        high_spatial_res: bool = False,
        high_time_res: bool = False,
        short_lifetime: bool = False,
        overwrite: bool = False,
        # model_params: Optional[Dict] = None,
    ) -> Dict[str, Dict]:
        """Reads footprints data files and returns the UUIDS of the Datasources
        the processed data has been assigned to

        Args:
            filepath: Path of file to load
            site: Site name
            height: Height above ground level in metres
            domain: Domain of footprints
            model: Model used to create footprint (e.g. NAME or FLEXPART)
            metmodel: Underlying meteorlogical model used (e.g. UKV)
            species: Species name. Only needed if footprint is for a specific species e.g. co2 (and not inert)
            network: Network name
            period: Period of measurements. Only needed if this can not be inferred from the time coords
            continuous: Whether time stamps have to be continuous.
            retrieve_met: Whether to also download meterological data for this footprints area
            high_spatial_res : Indicate footprints include both a low and high spatial resolution.
            high_time_res: Indicate footprints are high time resolution (include H_back dimension)
                           Note this will be set to True automatically if species="co2" (Carbon Dioxide).
            short_lifetime: Indicate footprint is for a short-lived species. Needs species input.
                            Note this will be set to True if species has an associated lifetime.
            overwrite: Overwrite any currently stored data

        Returns:
            dict: UUIDs of Datasources data has been assigned to
        """
        from openghg.store import assign_data, datasource_lookup, infer_date_range, load_metastore
        from openghg.util import clean_string, hash_file, species_lifetime, timestamp_now
        from xarray import open_dataset

        filepath = Path(filepath)

        site = clean_string(site)
        network = clean_string(network)
        height = clean_string(height)
        domain = clean_string(domain)

        fp = Footprints.load()

        # Load in the metadata store
        metastore = load_metastore(key=fp._metakey)

        file_hash = hash_file(filepath=filepath)
        if file_hash in fp._file_hashes and not overwrite:
            print(
                f"This file has been uploaded previously with the filename : {fp._file_hashes[file_hash]} - skipping."
            )

        fp_data = open_dataset(filepath)

        if species == "co2":
            # Expect co2 data to have high time resolution
            if not high_time_res:
                print("Updating high_time_res to True for co2 data")
                high_time_res = True

        if short_lifetime and not species:
            raise ValueError(
                "When indicating footprint is for short lived species, 'species' input must be included"
            )
        elif not short_lifetime and species:
            lifetime = species_lifetime(species)
            if lifetime is not None:
                # TODO: May want to add a check on length of lifetime here
                print("Updating short_lifetime to True since species has an associated lifetime")
                short_lifetime = True

        # Checking against expected format for footprints
        # Based on configuration (some user defined, some inferred)
        Footprints.validate_data(
            fp_data,
            high_spatial_res=high_spatial_res,
            high_time_res=high_time_res,
            short_lifetime=short_lifetime,
        )

        # Need to read the metadata from the footprints and then store it
        # Do we need to chunk the footprints / will a Datasource store it correctly?
        metadata: Dict[str, Union[str, float, List[float]]] = {}

        metadata["data_type"] = "footprints"
        metadata["site"] = site
        metadata["height"] = height
        metadata["domain"] = domain
        metadata["model"] = model

        if species is not None:
            metadata["species"] = clean_string(species)

        if network is not None:
            metadata["network"] = clean_string(network)

        if metmodel is not None:
            metadata["metmodel"] = clean_string(metmodel)

        fp_time = fp_data.time

        start_date, end_date, period_str = infer_date_range(
            fp_time, filepath=filepath, period=period, continuous=continuous
        )

        metadata["start_date"] = str(start_date)
        metadata["end_date"] = str(end_date)
        metadata["time_period"] = period_str

        metadata["max_longitude"] = round(float(fp_data["lon"].max()), 5)
        metadata["min_longitude"] = round(float(fp_data["lon"].min()), 5)
        metadata["max_latitude"] = round(float(fp_data["lat"].max()), 5)
        metadata["min_latitude"] = round(float(fp_data["lat"].min()), 5)

        if high_spatial_res:
            try:
                metadata["max_longitude_high"] = round(float(fp_data["lon_high"].max()), 5)
                metadata["min_longitude_high"] = round(float(fp_data["lon_high"].min()), 5)
                metadata["max_latitude_high"] = round(float(fp_data["lat_high"].max()), 5)
                metadata["min_latitude_high"] = round(float(fp_data["lat_high"].min()), 5)

                metadata["spatial_resolution"] = "high_spatial_resolution"
            except KeyError:
                raise KeyError("Expected high spatial resolution. Unable to find lat_high or lon_high data.")
        else:
            metadata["spatial_resolution"] = "standard_spatial_resolution"

        if high_time_res:
            metadata["time_resolution"] = "high_time_resolution"
        else:
            metadata["time_resolution"] = "standard_time_resolution"

        metadata["heights"] = [float(h) for h in fp_data.height.values]
        # Do we also need to save all the variables we have available in this footprints?
        metadata["variables"] = list(fp_data.data_vars)

        # if model_params is not None:
        #     metadata["model_parameters"] = model_params

        # Set the attributes of this Dataset
        fp_data.attrs = {"author": "OpenGHG Cloud", "processed": str(timestamp_now())}

        # This might seem longwinded now but will help when we want to read
        # more than one footprints at a time
        key = "_".join((site, domain, model, height))

        footprint_data: DefaultDict[str, Dict[str, Union[Dict, Dataset]]] = defaultdict(dict)
        footprint_data[key]["data"] = fp_data
        footprint_data[key]["metadata"] = metadata

        # These are the keys we will take from the metadata to search the
        # metadata store for a Datasource, they should provide as much detail as possible
        # to uniquely identify a Datasource
        required = ("site", "model", "height", "domain")
        lookup_results = datasource_lookup(metastore=metastore, data=footprint_data, required_keys=required)

        data_type = "footprints"
        datasource_uuids: Dict[str, Dict] = assign_data(
            data_dict=footprint_data,
            lookup_results=lookup_results,
            overwrite=overwrite,
            data_type=data_type,
        )

        fp.add_datasources(uuids=datasource_uuids, data=footprint_data, metastore=metastore)

        # Record the file hash in case we see this file again
        fp._file_hashes[file_hash] = filepath.name

        fp.save()

        metastore.close()

        return datasource_uuids

    @staticmethod
    def schema(
        particle_locations: bool = True,
        high_spatial_res: bool = False,
        high_time_res: bool = False,
        short_lifetime: bool = False,
    ) -> DataSchema:
        """
        Define schema for footprint Dataset.

        The returned schema depends on what the footprint represents,
        indicated using the keywords.
        By default, this will include "fp" variable but this will be superceded
        if high_spatial_res or high_time_res are specified.

        Args:
            particle_locations: Include 4-directional particle location variables:
                - "particle_location_[nesw]"
                and include associated additional dimensions ("height")
            high_spatial_res : Set footprint variables include high and low resolution options:
                - "fp_low"
                - "fp_high"
                and include associated additional dimensions ("lat_high", "lon_high").
            high_time_res : Set footprint variable to be high time resolution
                - "fp_HiTRes"
                and include associated dimensions ("H_back").
            short_lifetime: Include additional particle age parameters for short lived species:
                - "mean_age_particles_[nesw]"
        """

        # Names of data variables and associated dimensions (as a tuple)
        data_vars: Dict[str, Tuple[str, ...]] = {}
        # Internal data types of data variables and coordinates
        dtypes = {
            "lat": np.floating,  # Covers np.float16, np.float32, np.float64 types
            "lon": np.floating,
            "time": np.datetime64,
        }

        if not high_time_res and not high_spatial_res:
            # Includes standard footprint variable
            data_vars["fp"] = ("time", "lat", "lon")
            dtypes["fp"] = np.floating

        if high_spatial_res:
            # Include options for high spatial resolution footprint
            # This includes footprint data on multiple resolutions

            data_vars["fp_low"] = ("time", "lat", "lon")
            data_vars["fp_high"] = ("time", "lat_high", "lon_high")

            dtypes["fp_low"] = np.floating
            dtypes["fp_high"] = np.floating

        if high_time_res:
            # Include options for high time resolution footprint (usually co2)
            # This includes a footprint data with an additional hourly back dimension
            data_vars["fp_HiTRes"] = ("time", "lat", "lon", "H_back")
            dtypes["fp_HiTRes"] = np.floating
            dtypes["H_back"] = np.number  # float or integer

        # Includes particle location directions - one for each regional boundary
        if particle_locations:
            data_vars["particle_locations_n"] = ("time", "lon", "height")
            data_vars["particle_locations_e"] = ("time", "lat", "height")
            data_vars["particle_locations_s"] = ("time", "lon", "height")
            data_vars["particle_locations_w"] = ("time", "lat", "height")

            dtypes["height"] = np.floating
            dtypes["particle_locations_n"] = np.floating
            dtypes["particle_locations_e"] = np.floating
            dtypes["particle_locations_s"] = np.floating
            dtypes["particle_locations_w"] = np.floating

        # TODO: Could also add check for meteorological + other data
        # "pressure", "wind_speed", "wind_direction", "PBLH"
        # "release_lon", "release_lat"

        # Include options for short lifetime footprints (short-lived species)
        # This includes additional particle ages (allow calculation of decay based on particle lifetimes)
        if short_lifetime:
            data_vars["mean_age_particles_n"] = ("time", "lon", "height")
            data_vars["mean_age_particles_e"] = ("time", "lat", "height")
            data_vars["mean_age_particles_s"] = ("time", "lon", "height")
            data_vars["mean_age_particles_w"] = ("time", "lat", "height")

            dtypes["mean_age_particles_n"] = np.floating
            dtypes["mean_age_particles_e"] = np.floating
            dtypes["mean_age_particles_s"] = np.floating
            dtypes["mean_age_particles_w"] = np.floating

        data_format = DataSchema(data_vars=data_vars, dtypes=dtypes)

        return data_format

    @staticmethod
    def validate_data(
        data: Dataset,
        particle_locations: bool = True,
        high_spatial_res: bool = False,
        high_time_res: bool = False,
        short_lifetime: bool = False,
    ) -> None:
        """
        Validate data against Footprint schema - definition from
        Footprints.schema(...) method.

        Args:
            data : xarray Dataset in expected format

            See Footprints.schema() method for details on optional inputs.

        Returns:
            None

            Raises a ValueError with details if the input data does not adhere
            to the Footprints schema.
        """
        data_schema = Footprints.schema(
            particle_locations=particle_locations,
            high_spatial_res=high_spatial_res,
            high_time_res=high_time_res,
            short_lifetime=short_lifetime,
        )
        data_schema.validate_data(data)

    def search(
        self,
        site: str,
        network: str,
        start_date: Optional[Union[str, Timestamp]],
        end_date: Optional[Union[str, Timestamp]],
    ) -> NoReturn:
        """Search for a footprints from a specific site and network, return a dictionary of data
        so the user can choose
        """
        raise NotImplementedError()

    def retrieve(self, uuid: str, dates: str) -> NoReturn:
        """"""
        raise NotImplementedError()

    def _get_metdata(self) -> NoReturn:
        """This retrieves the metadata for this footprints"""
        raise NotImplementedError()



__all__ = ["JobCreator"]

# type: ignore


class JobCreator:
    """Create a job to be run on a HPC cluster

    WIP: Skeleton at the moment, will be filled out for use with multiple HPC
    resources/CitC
    """

    def __init__(self):
        self._jobs = []

    def create_job(self, name, requirements, data):
        """Create a job script

        TODO: Create an actual job script for

        For now just create a script file

        Args:
            name (str): Job name
            requirements (dict): Dictionary containing requested resources
            Example:
                requirements = {"cores": 16, "memory": 128G, "duration": 12h}
            For a job running with 16 cores and requesting 128 GB of memory for 12 hours
            (if duration is required)
        Returns:
            str: Job script as string
        """
        return "Sweet job script"



from datetime import datetime
from pathlib import Path
from typing import Dict

from openghg.standardise.meta import assign_attributes
from openghg.types import pathType
from openghg.util import clean_string, load_json
from pandas import NaT, read_csv


def parse_npl(
    data_filepath: pathType,
    site: str = "NPL",
    network: str = "LGHG",
    inlet: str = None,
    instrument: str = None,
    sampling_period: str = None,
    measurement_type: str = None,
) -> Dict:
    """Reads NPL data files and returns the UUIDS of the Datasources
    the processed data has been assigned to

    Args:
        data_filepath: Path of file to load
        site: Site name
    Returns:
        list: UUIDs of Datasources data has been assigned to
    """

    if sampling_period is None:
        sampling_period = "NOT_SET"

    data_filepath = Path(data_filepath)

    site = "NPL"

    attributes_data = load_json(filename="attributes.json")
    npl_params = attributes_data["NPL"]

    # mypy doesn't like NaT or NaNs - look into this
    def parser(date: str):  # type: ignore
        try:
            return datetime.strptime(str(date), "%d/%m/%Y %H:%M")
        except ValueError:
            return NaT

    data = read_csv(data_filepath, index_col=0, date_parser=parser)

    # Drop the NaT/NaNs
    data = data.loc[data.index.dropna()]

    # Rename columns
    rename_dict = {"Cal_CO2_dry": "CO2", "Cal_CH4_dry": "CH4"}

    data = data.rename(columns=rename_dict)
    data.index.name = "time"

    if inlet is None:
        inlet = "NA"

    gas_data = {}
    for species in data.columns:
        processed_data = data.loc[:, [species]].sort_index().to_xarray()

        # Convert methane to ppb
        if species == "CH4":
            processed_data[species] *= 1000

        # No averaging applied to raw obs, set variability to 0 to allow get_obs to calculate
        # when averaging
        processed_data["{} variability".format(species)] = processed_data[species] * 0.0

        site_attributes = npl_params["global_attributes"]
        site_attributes["inlet_height_magl"] = npl_params["inlet"]
        site_attributes["instrument"] = npl_params["instrument"]

        metadata = {
            "species": clean_string(species),
            "sampling_period": str(sampling_period),
            "site": "NPL",
            "network": "LGHG",
            "inlet": inlet,
        }

        # TODO - add in better metadata reading
        gas_data[species] = {
            "metadata": metadata,
            "data": processed_data,
            "attributes": site_attributes,
        }

    gas_data = assign_attributes(data=gas_data, site=site, network=network)

    return gas_data



from typing import List, Optional, Union

import plotly.graph_objects as go
from openghg.dataobjects import ObsData


def plot_timeseries(
    data: Union[ObsData, List[ObsData]],
    title: Optional[str] = None,
    xvar: Optional[str] = None,
    yvar: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    units: Optional[str] = None,
) -> go.Figure:
    """Plot a timeseries

    Args:
        data: ObsData object or list of objects
        title: Title for figure
        xvar: x axis variable, defaults to time
        yvar: y axis variable, defaults to species
        xlabel: Label for x axis
        ylabel: Label for y axis
        units: Units for y axis
    Returns:
        go.Figure: Plotly Graph Object Figure
    """
    from openghg.util import load_json

    if not isinstance(data, list):
        data = [data]

    font = {"size": 14}
    title_layout = {"text": title, "y": 0.9, "x": 0.5, "xanchor": "center", "yanchor": "top"}

    if title is None:
        title = ""

    if xlabel is None:
        xlabel = "Date"

    if ylabel is None:
        ylabel = "Concentration"

    if units is not None:
        ylabel += f"  ({units})"

    layout = go.Layout(
        title=title_layout,
        xaxis=dict(title=xlabel),
        yaxis=dict(title=ylabel),
        font=font,
    )

    # Create a single figure
    fig = go.Figure(layout=layout)

    for i, to_plot in enumerate(data):
        metadata = to_plot.metadata
        dataset = to_plot.data

        species = metadata["species"]
        site = metadata["site"]
        inlet = metadata["inlet"]

        legend = f"{site} - {species} - {inlet}"

        if xvar is not None:
            x_data = dataset[xvar]
        else:
            x_data = dataset.time

        if yvar is not None:
            y_data = dataset[yvar]
        else:
            try:
                y_data = dataset[species]
            except KeyError:
                y_data = dataset["mf"]

        if units is not None or len(data) > 1:

            data_attrs = y_data.attrs
            data_units = data_attrs.get("units", "1")

            if i == 0:
                if units:
                    attributes_data = load_json("attributes.json")
                    unit_interpret = attributes_data["unit_interpret"]
                    unit_value = unit_interpret.get(units, "1")
                else:
                    unit_value = data_units

            unit_conversion = float(data_units) / float(unit_value)
        else:
            unit_conversion = 1

        y_data *= unit_conversion

        fig.add_trace(go.Scatter(name=legend, x=x_data, y=y_data, mode="lines"))

    return fig


# def plot_timeseries(
#     data: Union[ObsData, List[ObsData]], xvar: str, yvar: str, xlabel: str, ylabel: str
# ) -> None:
#     """Plot timeseries data using Plotly

#     Args:
#         data: One or more ObsData objects
#         xvar: Variable to plot on x axis
#         yvar: Variable to plot on y axis
#         xlabel: Label for x axis
#         ylabel: Label for y axis
#     Returns:
#         plot type
#     """


#     for dat in data:



import logging
import os as _os
import sys as _sys

from . import (
    analyse,
    cloud,
    dataobjects,
    objectstore,
    retrieve,
    service,
    standardise,
    store,
    types,
    util,
)
from ._version import get_versions  # type: ignore

__all__ = [
    "analyse",
    "cloud",
    "dataobjects",
    "objectstore",
    "retrieve",
    "service",
    "standardise",
    "store",
    "util",
    "types",
]

if _sys.version_info.major < 3:
    raise ImportError("openghg requires Python >= 3.8")

if _sys.version_info.minor < 8:
    raise ImportError("openghg requires Python >= 3.8")

# Let's do some quick checks for required environment variables
_cloud = _os.environ.get("OPENGHG_CLOUD", False)
_hub = _os.environ.get("OPENGHG_HUB", False)
_openghg_path = _os.environ.get("OPENGHG_PATH", False)

if not (_cloud or _hub):
    if not _openghg_path:
        raise ValueError(
            "No environment variable OPENGHG_PATH found, please set to use the local object store"
        )

del _cloud, _hub, _openghg_path

v = get_versions()

__version__ = v.get("version")
__branch__ = v.get("branch")
__repository__ = v.get("repository")
__revisionid__ = v.get("full-revisionid")

del v, get_versions

import os as _os
from pathlib import Path as _Path

cloud_env = _os.environ.get("OPENGHG_CLOUD", False)
hub_env = _os.environ.get("OPENGHG_HUB", False)

if cloud_env or hub_env:
    logfile_path = "/tmp/openghg.log"
else:
    logfile_path = str(_Path.home().joinpath("openghg.log"))

# Start module level logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Create file handler for log file - set to DEBUG (maximum detail)
fileHandler = logging.FileHandler(logfile_path)  # May want to update this to user area
fileFormatter = logging.Formatter("%(asctime)s:%(levelname)s:%(name)s:%(message)s")
fileHandler.setFormatter(fileFormatter)
fileHandler.setLevel(logging.DEBUG)
logger.addHandler(fileHandler)

# Create console handler - set to WARNING (lower level)
consoleHandler = logging.StreamHandler()
consoleFormatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
consoleHandler.setFormatter(consoleFormatter)
consoleHandler.setLevel(logging.WARNING)
logger.addHandler(consoleHandler)

del logfile_path, hub_env, cloud_env



import math
from copy import deepcopy
from typing import Dict, List, Optional

from openghg.util import is_number


def metadata_default_keys() -> List:
    """
    Define default values expected within ObsSurface metadata
    """
    default_keys = [
        "site",
        "species",
        "inlet",
        "inlet_height_magl",
        "network",
        "instrument",
        "sampling_period",
        "calibration_scale",
        "data_owner",
        "data_owner_email",
        "station_longitude",
        "station_latitude",
        "station_long_name",
        "station_height_masl",
    ]

    return default_keys


def sync_surface_metadata(
    metadata: Dict,
    attributes: Dict,
    keys_to_add: Optional[List] = None,
) -> Dict:
    """Makes sure any duplicated keys between the metadata and attributes
    dictionaries match and that certain keys are present in the metadata.

    Args:
        metadata: Dictionary of metadata
        attributes: Attributes
        keys_to_add: Add these keys to the metadata, if not present, based on
        the attribute values. Note: this skips any keys which can't be
        copied from the attribute values.
    Returns:
        dict: Copy of metadata updated with attributes
    """
    meta_copy = deepcopy(metadata)

    # Check if we have differences
    for key, value in metadata.items():
        try:
            attr_value = attributes[key]

            # This should mainly be used for lat/long
            relative_tolerance = 1e-3

            if is_number(attr_value):
                if not math.isclose(float(attr_value), float(value), rel_tol=relative_tolerance):
                    raise ValueError(
                        f"Value not within tolerance, metadata: {value} - attributes: {attr_value}"
                    )
            else:
                if str(value).lower() != str(attr_value).lower():
                    raise ValueError(
                        f"Metadata mismatch for '{key}', metadata: {value} - attributes: {attr_value}"
                    )
        except KeyError:
            # Key wasn't in attributes for comparison
            pass

    default_keys_to_add = metadata_default_keys()

    if keys_to_add is None:
        keys_to_add = default_keys_to_add

    # Check set of keys which should be in metadata and add if not present
    for key in keys_to_add:
        if key not in meta_copy.keys():
            try:
                meta_copy[key] = attributes[key]
            except KeyError:
                print(f"WARNING: {key} key not in attributes or metadata")

    return meta_copy



from ._boundary_conditions import BoundaryConditions
from ._data_schema import DataSchema
from ._emissions import Emissions
from ._eulerian_model import EulerianModel
from ._footprints import Footprints
from ._infer_time import infer_date_range
from ._metadata import ObjectStorage, datasource_lookup, load_metastore, metastore_manager
from ._metstore import METStore
from ._obsmobile import ObsMobile
from ._obssurface import ObsSurface
from ._populate import add_noaa_obspack
from ._rank import rank_sources
from ._recombination import recombine_datasets, recombine_multisite
from ._segment import assign_data



import base64
import datetime


def datetime_to_datetime(d: datetime.datetime) -> datetime.datetime:
    """Return the passed datetime as a datetime that is clean
    and usable by Acquire. This will move the datetime to UTC,
    adding the timezone if this is missing

    Args:
         d (datetime): datetime to convert to UTC
    Returns:
         datetime: UTC datetime useable by Acquire
    """
    if not isinstance(d, datetime.datetime):
        raise TypeError(f"The passed object {str(d)} is not a valid datetime")

    if d.tzinfo is None:
        return d.replace(tzinfo=datetime.timezone.utc)
    else:
        return d.astimezone(datetime.timezone.utc)


def get_datetime_now():
    """Return the current time in the UTC timezone. This creates an
    object that will be properly stored using datetime_to_string
    and string_to_datetime

    Returns:
         datetime: Current datetime
    """
    return datetime.datetime.now(tz=datetime.timezone.utc)


def string_to_bytes(s: str) -> bytes:
    """Return the passed base64 utf-8 encoded binary data
    back converted from a string back to bytes. Note that
    this can only convert strings that were encoded using
    bytes_to_string - you cannot use this to convert
    arbitrary strings to bytes

    Args:
        s: base64 byte object to decode
    Returns:
        bytes: bytes
    """
    if s is None:
        return None
    else:
        return base64.b64decode(s.encode("utf-8"))


def bytes_to_string(b: bytes) -> str:
    """Return the passed binary bytes safely encoded to
    a base64 utf-8 string

    Args:
        b: binary bytes to encode
    Returns:
        str: UTF-8 encoded string
    """
    if b is None:
        return None
    else:
        return base64.b64encode(b).decode("utf-8")



from dataclasses import dataclass

from ._basedata import _BaseData

__all__ = ["FootprintData"]


@dataclass(frozen=True)
class FootprintData(_BaseData):
    """This class is used to return observations data from the get_footprint function

    Args:
        data: xarray Dataframe
        metadata: Dictionary of metadata including model run parameters
        flux: Dictionary of flux data
        bc: Boundary conditions dictionary
        species: Species name
        scales: Measurements scale
        units: Measurements units
    """

    def __str__(self) -> str:
        return f"Data: {self.data}\nMetadata : {self.metadata}"



from collections import defaultdict
from typing import DefaultDict, Dict, Optional

from openghg.util import (
    compress,
    compress_json,
    compress_str,
    decompress,
    decompress_json,
    hash_bytes,
)


def unpackage(data: Dict) -> Dict:
    """Unpackages and checks a dictionary created by the package_from_function function.
    This checks the SHA1 sums and decompresses the data.

    Args:
        data: Dictionary
    Returns:
        dict: Dictionary containing data and metadata if given
    """
    unpacked = {}

    file_metadata = data["file_metadata"]
    remote_sum = file_metadata["data"]["sha1_hash"]

    decompressed_data = decompress(data=data["data"])
    local_sum = hash_bytes(data=decompressed_data)

    unpacked["data"] = decompressed_data

    if not remote_sum == local_sum:
        raise ValueError(f"Hash mismatch, remote {remote_sum} - local {local_sum}.")

    try:
        compressed_metadata = data["metadata"]
    except KeyError:
        pass
    else:
        unpacked["metadata"] = decompress_json(data=compressed_metadata)

    return unpacked


def package_from_function(data: bytes, metadata: Optional[str] = None) -> Dict:
    """Creates a package of data ready to be sent back to the caller.
    This calculates the SHA1 sum of the passed data and compresses it.
    If metadata is passed this is added to the returned dictionary. No SHA1
    is calculated for the metadata.

    NOTE: This function should only be used internally by a serverless function.

    Args:
        data: Binary data
        metadata: Result of json.dumps
    Returns:
        dict: Dictionary of compressed data and file metadata.
    """
    sha1_hash = hash_bytes(data=data)
    compressed_data = compress(data=data)
    compression_type = "bz2"

    packaged: DefaultDict = defaultdict(dict)
    packaged["found"] = True
    packaged["data"] = compressed_data
    packaged["file_metadata"]["data"] = {"sha1_hash": sha1_hash, "compression_type": compression_type}

    if metadata is not None:
        try:
            compressed_metadata = compress_str(s=metadata)
        except AttributeError:
            try:
                compressed_metadata = compress_json(data=metadata)
            except Exception as e:
                raise TypeError(f"Unable to process this object: {e}")

        packaged["metadata"] = compressed_metadata
        packaged["file_metadata"]["metadata"] = {"sha1_hash": False, "compression_type": compression_type}

    return dict(packaged)



from typing import Any, Dict, List, Optional, Union

from openghg.dataobjects import ObsData
from openghg.util import running_on_hub


def retrieve_atmospheric(
    site: str,
    species: Optional[Union[str, List]] = None,
    sampling_height: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    force_retrieval: bool = False,
    data_level: int = 2,
) -> Union[ObsData, List[ObsData], None]:
    """Retrieve ICOS atmospheric measurement data. If data is found in the object store it is returned. Otherwise
    data will be retrieved from the ICOS Carbon Portal. Data retrieval from the Carbon Portal may take a short time.
    If only a single data source is found an ObsData object is returned, if multiple a list of ObsData objects
    if returned, if nothing then None.

    Args:
        site: Site code
        species: Species name
        start_date: Start date
        end_date: End date
        force_retrieval: Force the retrieval of data from the ICOS Carbon Portal
        data_level: ICOS data level (1, 2)
        - Data level 1: Near Real Time Data (NRT) or Internal Work data (IW).
        - Data level 2: The final quality checked ICOS RI data set, published by the CFs,
                        to be distributed through the Carbon Portal.
                        This level is the ICOS-data product and free available for users.
        See https://icos-carbon-portal.github.io/pylib/modules/#stationdatalevelnone
        bypass_call: Bypass the remote function call, used to shortcut calls within a the serverless
        function call environment.
    Returns:
        ObsData, list[ObsData] or None
    """
    return retrieve(
        site=site,
        species=species,
        sampling_height=sampling_height,
        start_date=start_date,
        end_date=end_date,
        force_retrieval=force_retrieval,
        data_level=data_level,
    )


def retrieve(**kwargs: Any) -> Union[ObsData, List[ObsData], None]:
    """Retrieve data from the ICOS Carbon Portal. If data is found in the local object store
    it will be retrieved from there first.

    This function detects the running environment and routes the call
    to either the cloud or local search function.

    Example / commonly used arguments are given below.

    Args:
        site: Site code
        species: Species name
        start_date: Start date
        end_date: End date
        force_retrieval: Force the retrieval of data from the ICOS Carbon Portal
        data_level: ICOS data level (1, 2)
        - Data level 1: Near Real Time Data (NRT) or Internal Work data (IW).
        - Data level 2: The final quality checked ICOS RI data set, published by the CFs,
                        to be distributed through the Carbon Portal.
                        This level is the ICOS-data product and free available for users.
        See https://icos-carbon-portal.github.io/pylib/modules/#stationdatalevelnone
        bypass_call: Bypass the remote function call, used to shortcut calls within a the serverless
        function call environment.
    Returns:
        ObsData, list[ObsData] or None
    """
    from io import BytesIO

    from openghg.cloud import call_function, unpackage
    from xarray import load_dataset

    # The hub is the only place we want to make remote calls
    if running_on_hub():
        post_data: Dict[str, Union[str, Dict]] = {}
        post_data["function"] = "retrieve_icos"
        post_data["search_terms"] = kwargs

        call_result = call_function(data=post_data)

        content = call_result["content"]
        found = content["found"]

        if not found:
            return None

        observations = content["data"]

        obs_data = []
        for package in observations.values():
            unpackaged = unpackage(data=package)
            buf = BytesIO(unpackaged["data"])
            ds = load_dataset(buf)
            obs = ObsData(data=ds, metadata=unpackaged["metadata"])

            obs_data.append(obs)

        if len(obs_data) == 1:
            return obs_data[0]
        else:
            return obs_data
    else:
        return local_retrieve(**kwargs)


def local_retrieve(
    site: str,
    species: Optional[Union[str, List]] = None,
    sampling_height: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    force_retrieval: bool = False,
    data_level: int = 2,
    **kwargs: Any,
) -> Union[ObsData, List[ObsData], None]:
    """Retrieve ICOS atmospheric measurement data. If data is found in the object store it is returned. Otherwise
    data will be retrieved from the ICOS Carbon Portal. Data retrieval from the Carbon Portal may take a short time.
    If only a single data source is found an ObsData object is returned, if multiple a list of ObsData objects
    if returned, if nothing then None.

    Args:
        site: Site code
        species: Species name
        start_date: Start date
        end_date: End date
        force_retrieval: Force the retrieval of data from the ICOS Carbon Portal
        data_level: ICOS data level (1, 2)
        - Data level 1: Near Real Time Data (NRT) or Internal Work data (IW).
        - Data level 2: The final quality checked ICOS RI data set, published by the CFs,
                        to be distributed through the Carbon Portal.
                        This level is the ICOS-data product and free available for users.
        See https://icos-carbon-portal.github.io/pylib/modules/#stationdatalevelnone
    Returns:
        ObsData, list[ObsData] or None
    """
    from openghg.dataobjects import ObsData
    from openghg.retrieve import search_surface
    from openghg.store import ObsSurface
    from openghg.util import to_lowercase

    if not 1 <= data_level <= 2:
        print("Error: data level must be 1 or 2.")

    # NOTE - we skip ranking here, will we be ranking ICOS data?
    results = search_surface(
        site=site,
        species=species,
        sampling_height=sampling_height,
        network="ICOS",
        data_source="icoscp",
        start_date=start_date,
        end_date=end_date,
        icos_data_level=data_level,
        skip_ranking=True,
    )

    if results and not force_retrieval:
        obs_data = results.retrieve_all()
    else:
        # We'll also need to check we have current data
        standardised_data = _retrieve_remote(site=site, species=species, data_level=data_level)

        if standardised_data is None:
            return None

        ObsSurface.store_data(data=standardised_data)

        # Create the expected ObsData type
        obs_data = []
        for data in standardised_data.values():
            measurement_data = data["data"]
            # These contain URLs that are case sensitive so skip lowercasing these
            skip_keys = ["citation_string", "instrument_data", "dobj_pid"]
            metadata = to_lowercase(data["metadata"], skip_keys=skip_keys)
            obs_data.append(ObsData(data=measurement_data, metadata=metadata))

    if isinstance(obs_data, list) and len(obs_data) == 1:
        return obs_data[0]
    else:
        return obs_data


def _retrieve_remote(
    site: str,
    data_level: int,
    species: Optional[Union[str, List]] = None,
    sampling_height: Optional[str] = None,
) -> Optional[Dict]:
    """Retrieve ICOS data from the ICOS Carbon Portal and standardise it into
    a format expected by OpenGHG. A dictionary of metadata and Datasets

    Args:
        site: ICOS site code, for site codes see
        https://www.icos-cp.eu/observations/atmosphere/stations
        data_level: ICOS data level (1, 2)
        - Data level 1: Near Real Time Data (NRT) or Internal Work data (IW).
        - Data level 2: The final quality checked ICOS RI data set, published by the CFs,
                        to be distributed through the Carbon Portal.
                        This level is the ICOS-data product and free available for users.
        See https://icos-carbon-portal.github.io/pylib/modules/#stationdatalevelnone
        species: Species name
        sampling_height: Sampling height in metres
    Returns:
        dict or None: Dictionary of processed data and metadata if found
    """
    # icoscp isn't available to conda so we've got to resort to this for now
    try:
        from icoscp.cpb.dobj import Dobj  # type: ignore
        from icoscp.station import station  # type: ignore
    except ImportError:
        raise ImportError(
            "Cannot import icoscp, if you've installed OpenGHG using conda please run: pip install icoscp"
        )

    import re

    from openghg.standardise.meta import assign_attributes
    from openghg.util import load_json
    from pandas import to_datetime

    if species is None:
        species = ["CO", "CO2", "CH4"]

    if not isinstance(species, list):
        species = [species]

    # We should first check if it's stored in the object store
    # Will need to make sure ObsSurface can accept the datasets we
    # create from the ICOS data
    stat = station.get(stationId=site.upper())

    if not stat.valid:
        print("Please check you have passed a valid ICOS site.")
        return None

    data_pids = stat.data(level=data_level)

    # We want to get the PIDs of the data for each species here
    species_upper = [s.upper() for s in species]
    # For this see https://stackoverflow.com/a/55335207
    search_str = r"\b(?:{})\b".format("|".join(map(re.escape, species_upper)))
    # Now filter the dataframe so we can extraxt the PIDS
    filtered_sources = data_pids[data_pids["specLabel"].str.contains(search_str)]

    if sampling_height is not None:
        sampling_height = str(float(sampling_height.rstrip("m")))
        filtered_sources = filtered_sources[
            [sampling_height in x for x in filtered_sources["samplingheight"]]
        ]

    if filtered_sources.empty:
        print(
            f"No sources found for {species} at {site}. Please check with the ICOS Carbon Portal that this data is available."
        )
        return None

    # Now extract the PIDs along with some data about them
    dobj_urls = filtered_sources["dobj"].tolist()

    site_metadata = load_json("icos_atmos_site_metadata.json")

    standardised_data: Dict[str, Dict] = {}

    for dobj_url in dobj_urls:
        dobj = Dobj(dobj_url)
        # We need to pull the data down as .info (metadata) is populated further on this step
        dataframe = dobj.get()
        # This is the metadata
        dobj_info = dobj.meta

        metadata = {}

        specific_info = dobj_info["specificInfo"]
        col_data = specific_info["columns"]

        for col in col_data:
            # Find the species
            for s in species:
                if col["label"] == s.lower():
                    measurement_type = col["valueType"]["self"]["label"].lower()
                    units = col["valueType"]["unit"].lower()
                    this_species = str(s)
                    break

        metadata["species"] = this_species
        acq_data = specific_info["acquisition"]
        station_data = acq_data["station"]

        to_store: Dict[str, Any] = {}
        try:
            instrument_metadata = acq_data["instrument"]
        except KeyError:
            to_store["instrument"] = "NA"
            to_store["instrument_data"] = "NA"
        else:
            # Do some tidying of the instrument metadata
            instruments = set()
            cleaned_instrument_metadata = []

            if not isinstance(instrument_metadata, list):
                instrument_metadata = [instrument_metadata]

            for inst in instrument_metadata:
                instrument_name = inst["label"]
                instruments.add(instrument_name)
                uri = inst["uri"]

                cleaned_instrument_metadata.extend([instrument_name, uri])

            if len(instruments) == 1:
                instrument = instruments.pop()
            else:
                instrument = "multiple"

            to_store["instrument"] = instrument
            to_store["instrument_data"] = cleaned_instrument_metadata

        metadata.update(to_store)

        metadata["site"] = station_data["id"]
        metadata["measurement_type"] = measurement_type
        metadata["units"] = units

        _sampling_height = acq_data["samplingHeight"]
        metadata["sampling_height"] = f"{int(float(_sampling_height))}m"
        metadata["inlet_height_magl"] = f"{int(float(_sampling_height))}m"
        metadata["sampling_height_units"] = "metres"
        metadata["inlet"] = f"{int(float(_sampling_height))}m"

        loc_data = station_data["location"]
        metadata["station_long_name"] = loc_data["label"]
        metadata["station_latitude"] = str(loc_data["lat"])
        metadata["station_longitude"] = str(loc_data["lon"])
        metadata["station_altitude"] = f"{int(float(loc_data['alt']))}m"

        site_specific = site_metadata[site.upper()]
        metadata["data_owner"] = f"{site_specific['firstName']} {site_specific['lastName']}"
        metadata["data_owner_email"] = site_specific["email"]
        metadata["station_height_masl"] = f"{int(float(site_specific['eas']))}m"

        metadata["citation_string"] = dobj_info["references"]["citationString"]
        metadata["licence_name"] = dobj_info["references"]["licence"]["name"]
        metadata["licence_info"] = dobj_info["references"]["licence"]["url"]

        # Add ICOS in directly here for now
        metadata["network"] = "ICOS"
        metadata["data_type"] = "timeseries"
        metadata["data_source"] = "icoscp"
        metadata["icos_data_level"] = str(data_level)

        dataframe.columns = [x.lower() for x in dataframe.columns]
        dataframe = dataframe.dropna(axis="index")

        if not dataframe.index.is_monotonic_increasing:
            dataframe = dataframe.sort_index()

        spec = metadata["species"]
        inlet = metadata["inlet"]

        rename_cols = {
            "stdev": spec + " variability",
            "nbpoints": spec + " number_of_observations",
        }

        # TODO - add this back in once we've merged the fixes in
        # Try and conver the flag / userflag column to str
        # possible_flag_cols = ("flag", "userflag")
        # flag_col = [x for x in dataframe.columns if x in possible_flag_cols]

        # PR328
        # if flag_col:
        #     flag_str = flag_col[0]
        #     dataframe = dataframe.astype({flag_str: str})

        dataframe = dataframe.rename(columns=rename_cols).set_index("timestamp")

        dataframe.index.name = "time"
        dataframe.index = to_datetime(dataframe.index, format="%Y-%m-%d %H:%M:%S")

        dataset = dataframe.to_xarray()
        dataset.attrs.update(metadata)

        # So there isn't an easy way of getting a hash of a Dataset, can we do something
        # simple here we can compare data that's being added? Then we'll be able to make sure
        # ObsSurface.store_data won't accept data it's already seen
        data_key = f"{spec}_{inlet}"
        # TODO - do we need both attributes and metadata here?
        standardised_data[data_key] = {
            "metadata": metadata,
            "data": dataset,
            "attributes": metadata,
        }

    # Make sure everything is CF compliant
    standardised_data = assign_attributes(data=standardised_data)

    return standardised_data


# def _read_site_metadata():
#     """ Read site metadata from object store, if it doesn't exist we'll
#     retrieve it from the ICOS CP and store it.

#     Returns:
#         dict: Dictionary of site data
#     """
#     from openghg.objectstore import get_bucket, get_object_from_json
#     from openghg.types import ObjectStoreError
#     from openghg.util import timestamp_now
# raise NotImplementedError
#     key = "metadata/icos_atmos_site_metadata"
#     bucket = get_bucket()

#     try:
#         data = get_object_from_json(bucket=bucket, key=key)
#     except ObjectStoreError:
#         # Retrieve and store
#         from icoscp import station
#         station_data = station.getIdList()
#         metadata = {d.id: dict(d) for _, d in df.iterrows()}



# import os

# from Acquire.Client import PAR

# # type: ignore


# class JobDrive:
#     """ " This is used to upload files to the cloud drive for use in a
#     HPC job

#     Args:
#         par (Acquire.Client.PAR): Pre-authenticated request for access to cloud drive
#         par_secret (str): Secret / password to access the PAR
#     """

#     def __init__(self, par, par_secret=None):
#         if not isinstance(par, PAR):
#             raise TypeError("par argument must be of type Acquire.Client.PAR")

#         self._par = par
#         self._par_secret = par_secret
#         self._drive = par.resolve(secret=par_secret)

#     def upload(self, files, directory="input"):
#         """Upload files to the cloud drive that's accessed using the Acquire
#         PAR

#         Args:
#             files (str, Path, list): File(s) to be uploaded
#             directory (str, default="input"): Name of directory in which to place files
#         Returns:
#             None
#         """
#         if not isinstance(files, list):
#             files = [files]

#         # 50 MB in bytes
#         chunk_limit = 50 * 1024 * 1024

#         # We might not have any data files to upload
#         for f in files:
#             filesize = os.path.getsize(f)
#             if filesize < chunk_limit:
#                 self._drive.upload(filename=f, directory=directory)
#             else:
#                 self._drive.chunk_upload(filename=f, directory=directory)

#     def list_files(self):
#         """List files in drive

#         Returns:
#             list: List of files
#         """
#         return self._drive.list_files(include_metadata=True)

#     def download(self, files, local_dir=None):
#         """Download the files in the given list to a local directory

#         Args:
#             files (list): File(s) on drive
#             local_dir (Path, str): Path of folder to download files to
#         Returns:
#             dict: Dictionary keyed by filename containing file metadata
#         """
#         if not isinstance(files, list):
#             files = [files]

#         # if local_dir is None:
#         #     date_str = datetime.now().strftime("%Y%m%d-%H%M%S")
#         #     folder_name = f"drive_download_{date_str}"
#         #     local_dir = Path.home().joinpath(folder_name).mkdir()

#         file_metadata = {}
#         for fname in files:
#             file_metadata[fname] = self._drive.download(filename=fname, directory=local_dir)

#         return file_metadata



from pathlib import Path
from typing import Dict, Optional, Union


def parse_openghg(
    filepath: Path,
    species: str,
    source: str,
    domain: str,
    date: Optional[str] = None,
    high_time_resolution: Optional[bool] = False,
    period: Optional[Union[str, tuple]] = None,
    continuous: bool = True,
) -> Dict:
    """
    Read and parse input emissions data already in OpenGHG format.

    Args:
        filepath: Path to data file
    Returns:
        dict: Dictionary of data
    """
    from openghg.standardise.meta import assign_flux_attributes
    from openghg.store import infer_date_range
    from openghg.util import timestamp_now
    from xarray import open_dataset

    em_data = open_dataset(filepath)

    # Some attributes are numpy types we can't serialise to JSON so convert them
    # to their native types here
    attrs = {}
    for key, value in em_data.attrs.items():
        try:
            attrs[key] = value.item()
        except AttributeError:
            attrs[key] = value

    author_name = "OpenGHG Cloud"
    em_data.attrs["author"] = author_name

    metadata = {}
    metadata.update(attrs)

    metadata["species"] = species
    metadata["domain"] = domain
    metadata["source"] = source
    metadata["date"] = date
    metadata["author"] = author_name
    metadata["processed"] = str(timestamp_now())

    # As emissions files handle things slightly differently we need to check the time values
    # more carefully.
    # e.g. a flux / emissions file could contain e.g. monthly data and be labelled as 2012 but
    # contain 12 time points labelled as 2012-01-01, 2012-02-01, etc.

    em_time = em_data.time

    start_date, end_date, period_str = infer_date_range(
        em_time, filepath=filepath, period=period, continuous=continuous
    )

    if date is None:
        # Check for how granular we should make the date label
        if "year" in period_str:
            date = f"{start_date.year}"
        elif "month" in period_str:
            date = f"{start_date.year}{start_date.month:02}"
        else:
            date = start_date.astype("datetime64[s]").astype(str)

    metadata["start_date"] = str(start_date)
    metadata["end_date"] = str(end_date)

    metadata["max_longitude"] = round(float(em_data["lon"].max()), 5)
    metadata["min_longitude"] = round(float(em_data["lon"].min()), 5)
    metadata["max_latitude"] = round(float(em_data["lat"].max()), 5)
    metadata["min_latitude"] = round(float(em_data["lat"].min()), 5)

    metadata["time_resolution"] = "high" if high_time_resolution else "standard"
    metadata["time_period"] = period_str

    key = "_".join((species, source, domain, date))

    emissions_data: Dict[str, dict] = {}
    emissions_data[key] = {}
    emissions_data[key]["data"] = em_data
    emissions_data[key]["metadata"] = metadata

    emissions_data = assign_flux_attributes(emissions_data)

    return emissions_data



import contextlib
import json
from typing import Dict, Iterator, Optional, Sequence

from openghg.objectstore import exists, get_bucket, get_object, set_object_from_json
from tinydb import Storage, TinyDB
from tinydb.middlewares import CachingMiddleware


@contextlib.contextmanager
def metastore_manager(key: str) -> Iterator:
    """A context manager to be used to open the metadata database
    at the given key.

    Args:
        key: Key for database in the object store
    """
    db = TinyDB(key, storage=CachingMiddleware(ObjectStorage))

    yield db

    db.close()


def load_metastore(key: str) -> TinyDB:
    """Load the metastore.

    Note: the database must be closed with db.close()
    to ensure correct writing of new values.

    Args:
        key: Key to metadata store
    Returns:
        TinyDB: instance of metadata database
    """
    return TinyDB(key, storage=CachingMiddleware(ObjectStorage))


class ObjectStorage(Storage):
    def __init__(self, key: str) -> None:
        self._key = key

    def read(self) -> Optional[Dict]:
        bucket = get_bucket()
        key = self._key

        if not exists(bucket=bucket, key=key):
            return None

        data = get_object(bucket=bucket, key=self._key)

        try:
            json_data: Dict = json.loads(data)
            return json_data
        except json.JSONDecodeError:
            return None

    def write(self, data: Dict) -> None:
        bucket = get_bucket()
        key = self._key

        set_object_from_json(bucket=bucket, key=key, data=data)

    def close(self) -> None:
        pass


def datasource_lookup(
    metastore: TinyDB, data: Dict, required_keys: Sequence[str], min_keys: Optional[int] = None
) -> Dict:
    """Search the metadata store for a Datasource UUID using the metadata in data. We expect the required_keys
    to be present and will require at leas min_keys of these to be present when searching.

    As some metadata value might change (such as data owners etc) we don't want to do an exact
    search on *all* the metadata so we extract a subset (the required keys) and search for these.

    Args:
        metastore: Metadata database
        data: Combined data dictionary of form {key: {data: Dataset, metadata: Dict}}
        required_keys: Iterable of keys to extract from metadata
        min_keys: The minimum number of required keys, if not given it will be set
        to the length of required_keys
    Return:
        dict: Dictionary of datasource information
    """
    from openghg.retrieve import metadata_lookup

    if min_keys is None:
        min_keys = len(required_keys)

    results = {}
    for key, _data in data.items():
        metadata = _data["metadata"]
        required_metadata = {k: v for k, v in metadata.items() if k in required_keys}

        if len(required_metadata) < min_keys:
            raise ValueError(
                f"The given metadata doesn't contain enough information, we need: {required_keys}"
            )

        results[key] = metadata_lookup(metadata=required_metadata, database=metastore)

    return results



import json
import tempfile
from datetime import datetime
from pathlib import Path

from openghg.jobs import SSHConnect
from openghg.util import get_datapath

# type: ignore


def run_job(username, hostname, password, job_data, known_host=False):
    """Set a job to run on a HPC service

    Args:
        username (str): Username for HPC cluster
        hostname (str): Hostname of HPC cluster
        password (str): Password to unlock private key used to access cluster
        job_data (dict): Data to run job
    Returns:
        dict: Dictionary of responses to commands executing when running the job
    """
    # Maybe this can be passed in / read from JSON depending on the service selected
    bc4_partitions = [
        "cpu_test",
        "dcv",
        "gpu",
        "gpu_veryshort",
        "hmem",
        "serial",
        "test",
        "veryshort",
    ]

    # These are used to write the Slurm script
    # This is a WIP, I might be reinventing the wheel here, check before doing any more
    name = job_data["name"]
    job_sched_command = job_data["job_sched_command"]
    partition = job_data["partition"]

    if partition not in bc4_partitions:
        raise ValueError(
            f"Invalid partition selected. Please select from one of the following :\n{bc4_partitions}"
        )

    n_nodes = job_data["n_nodes"]
    n_tasks_per_node = job_data["n_tasks_per_node"]
    n_cpus_per_task = job_data["n_cpus_per_task"]
    # This shuold be in GB
    memory_req = job_data["memory_req"]

    if not memory_req.endswith("G"):
        raise ValueError("Memory requirements must be in gigabytes and end with a G i.e. 128G")

    job_duration = job_data["job_duration"]

    # Create a dated and named filename and path in the user's home dir
    date_str = datetime.now().strftime("%Y%m%d-%H%M%S")
    name_date = f"{name}_{date_str}"
    # Name for the Slurm job script
    script_filename = f"run_job_{name_date}.sh"

    job_name = f"job_{name_date}"

    # Create a JSON file that will hold the job parameters
    json_dict = {}
    json_dict["job_name"] = job_name
    json_dict["job_data"] = job_data
    json_dict["script_filename"] = script_filename
    json_dict["par"] = job_data["par"]
    json_dict["par_secret"] = job_data["par_secret"]
    json_dict["run_command"] = job_data["run_command"]

    try:
        json_dict["compilation_command"] = job_data["compilation_command"]
    except KeyError:
        pass

    # Name of file to write JSON data to for transfer to server
    json_filename = f"job_data_{name_date}.json"

    # Here we want to write the jobscript and job parameters to file before transferring
    # to the cluster
    with tempfile.TemporaryDirectory() as tmpdir, SSHConnect() as sc:
        jobscript_path = Path(tmpdir).joinpath(script_filename)
        json_path = Path(tmpdir).joinpath(json_filename)

        with open(json_path, "w") as jf:
            json.dump(obj=json_dict, fp=jf, indent=4)

        with open(jobscript_path, "w") as rsh:
            rsh.write(
                f"""#!/bin/bash
        #SBATCH --partition={partition}
        #SBATCH --nodes={n_nodes}
        #SBATCH --ntasks-per-node={n_tasks_per_node}
        #SBATCH --cpus-per-task={n_cpus_per_task}
        #SBATCH --time={job_duration}
        #SBATCH --mem={memory_req}
        {job_sched_command}
            """
            )

        # Here we'll only copy the files we've created
        # Other input files will be copied from the cloud drive by the  script we're passing
        job_controller_path = get_datapath(filename="bc4_template.py", directory="job_controllers")

        files = [jobscript_path, json_path, job_controller_path]

        # TODO - I feel this shouldn't be hardwired, it probably won't change in the Docker image, but ?
        keypath = "/home/fnuser/.ssh/runner_key"

        sc.connect(
            username=username,
            hostname=hostname,
            keypath=keypath,
            password=password,
            known_host=known_host,
        )
        # sc.write_files(files=files, remote_dir="first_job")
        sc.write_files(files=files, remote_dir=job_name)

        response_list = sc.run_command(commands=f"cd {job_name}; python3 bc4_template.py {json_filename} &")

    return response_list



from pathlib import Path
from typing import Dict, Optional, Union


def parse_icos(
    data_filepath: Union[str, Path],
    site: str,
    inlet: str,
    instrument: str,
    network: str = "ICOS",
    sampling_period: Optional[str] = None,
    measurement_type: Optional[str] = None,
    header_type: str = "large",
) -> Dict:
    """Parses an ICOS data file and creates a dictionary containing the Dataset and metadata

    Args:
        data_filepath: Path to file
        site: Three letter site code
        network: Network name
        inlet: Inlet height
        instrument: Instrument name
        sampling_period: Sampling period e.g. 2 hour: 2H, 2 minute: 2m
        measurement_type: Measurement type e.g. insitu, flask
        header_type: ICOS data file with large (40 line) header or shorter single line header
            Options: large, small
    Returns:
        dict: Dictionary of gas data
    """
    from pathlib import Path

    from openghg.standardise.meta import assign_attributes
    from openghg.util import clean_string

    site = clean_string(site)
    inlet = clean_string(inlet)
    instrument = clean_string(instrument)
    network = clean_string(network)
    sampling_period = clean_string(sampling_period)
    measurement_type = clean_string(measurement_type)

    if not isinstance(data_filepath, Path):
        data_filepath = Path(data_filepath)

    if header_type == "large":
        gas_data = _read_data_large_header(
            data_filepath=data_filepath,
            site=site,
            inlet=inlet,
            network=network,
            instrument=instrument,
            sampling_period=sampling_period,
            measurement_type=measurement_type,
        )
    else:
        gas_data = _read_data_small_header(
            data_filepath=data_filepath,
            site=site,
            inlet=inlet,
            network=network,
            instrument=instrument,
            sampling_period=sampling_period,
            measurement_type=measurement_type,
        )

    # Ensure the data is CF compliant
    gas_data = assign_attributes(data=gas_data, site=site, sampling_period=sampling_period)

    return gas_data


def _read_data_large_header(
    data_filepath: Path,
    site: str,
    inlet: str,
    network: str,
    instrument: str,
    sampling_period: Optional[str] = None,
    measurement_type: Optional[str] = None,
) -> Dict:
    """Parses ICOS files with the larger (~40) line header

    Args:
        data_filepath: Path to file
        site: Three letter site code
        network: Network name
        inlet: Inlet height
        instrument: Instrument name
        sampling_period: Sampling period e.g. 2 hour: 2H, 2 minute: 2m
        measurement_type: Measurement type e.g. insitu, flask
    Returns:
        dict: Dictionary of gas data
    """
    from openghg.util import read_header
    from pandas import read_csv, to_datetime

    # Read metadata from the filename and cross check to make sure the passed
    # arguments match
    split_filename = data_filepath.name.split(".")

    try:
        site_fname = split_filename[0]
        species_fname = split_filename[1]
        file_sampling_period = split_filename[2]
        instrument_fname = split_filename[3]
        inlet_height_fname = split_filename[4]
    except IndexError:
        raise ValueError(
            "Unable to read metadata from filename. We expect a filename such as mhd.ch4.hourly.g2401.15m.dat"
        )

    if site_fname.lower() != site:
        raise ValueError("Site mismatch between site argument passed and filename.")

    if inlet is not None and inlet_height_fname.lower() != inlet:
        raise ValueError("Mismatch between inlet height passed and in filename.")

    if instrument is not None and instrument_fname.lower() != instrument:
        raise ValueError("Mismatch between instrument passed and that in filename.")

    # Read the header and check its length
    header = read_header(filepath=data_filepath)
    len_header = len(header)

    if len_header != 40:
        print(
            f"WARNING: We expect a header length of 40 but got {len_header}, \
                note that some metadata may not be collected, \
                please raise an issue on GitHub if this file format is to be expected."
        )

    dtypes = {
        "#Site": "string",
        "SamplingHeight": "string",
        "DecimalDate": "float",
        "Stdev": "float",
        "NbPoints": "int",
        "Flag": "string",
        "InstrumentId": "int",
        "QualityId": "string",
        "InternalFlag": "string",
        "AutoDescriptiveFlag": "string",
        "ManualDescriptiveFlag": "string",
    }

    df = read_csv(
        data_filepath,
        header=len_header - 1,
        sep=";",
        parse_dates={"time": [2, 3, 4, 5, 6]},
        index_col="time",
        na_values=["-9.990", "-999.990"],
        dtype=dtypes,
    )

    df.index = to_datetime(df.index, format="%Y %m %d %H %M")

    # Lowercase all the column titles
    df.columns = [str(c).lower() for c in df.columns]

    # Read some metadata before dropping the columns
    # sampling_height_data = df["samplingheight"][0]
    site_name_data = df["#site"][0]
    species_name_data = df.columns[3]

    if site != site_name_data.lower():
        raise ValueError("Site mismatch between site argument passed and site in data.")

    if species_fname != species_name_data.lower():
        raise ValueError("Speices mismatch between site passed and species in data.")

    # Drop the columns we don't want
    cols_to_keep = [species_name_data, "stdev", "nbpoints", "flag", "instrumentid"]
    df = df[cols_to_keep]

    # Remove rows with NaNs in the species or stdev columns
    df = df.dropna(axis="rows", subset=[species_name_data, "stdev"])

    # Drop duplicate indices
    df = df.loc[~df.index.duplicated(keep="first")]

    # Check if the index is sorted
    if not df.index.is_monotonic_increasing:
        df = df.sort_index()

    rename_dict = {
        "stdev": species_name_data + " variability",
        "nbpoints": species_name_data + " number_of_observations",
    }

    df = df.rename(columns=rename_dict)

    # Convert to xarray Dataset
    data = df.to_xarray()

    if file_sampling_period == "1minute":
        file_sampling_period = "60.0"
    elif file_sampling_period == "hourly":
        file_sampling_period = "3600.0"

    if sampling_period is not None:
        if file_sampling_period != sampling_period:
            raise ValueError("Mismatch between sampling period read from filename and that passed.")
    else:
        sampling_period = file_sampling_period

    metadata = {
        "site": site,
        "species": species_name_data,
        "inlet": inlet_height_fname,
        "sampling_period": file_sampling_period,
        "network": network,
        "instrument": instrument_fname,
    }

    if measurement_type is not None:
        metadata["measurement_type"] = measurement_type

    unit_line = header[22]
    if "MEASUREMENT UNIT" in unit_line:
        units = unit_line.split(":")[1].lower().strip()
        metadata["units"] = units

    scale_line = header[26]
    if "MEASUREMENT SCALE" in scale_line:
        calibration_scale = scale_line.split(":")[1].lower().lstrip(" ").replace(" ", "_").strip()
        metadata["calibration_scale"] = calibration_scale

    data_owner_line = header[18]
    if "CONTACT POINT" in data_owner_line:
        data_owner_email = data_owner_line.split(":")[1].split(",")[1].strip()
        metadata["data_owner_email"] = data_owner_email

    species_data = {
        species_name_data: {
            "metadata": metadata,
            "data": data,
        }
    }

    return species_data


def _read_data_small_header(
    data_filepath: Path,
    site: str,
    inlet: str,
    network: str,
    instrument: str,
    sampling_period: Optional[str] = None,
    measurement_type: Optional[str] = None,
) -> Dict:
    """Parses ICOS files with a single line header

    Args:
        data_filepath: Path to file
        site: Three letter site code
        network: Network name
        inlet: Inlet height
        instrument: Instrument name
        sampling_period: Sampling period e.g. 2 hour: 2H, 2 minute: 2m
        measurement_type: Measurement type e.g. insitu, flask
    Returns:
        dict: Dictionary of gas data
    """
    from openghg.util import read_header
    from pandas import Timestamp, read_csv

    # Read some metadata from the filename
    split_filename = data_filepath.name.split(".")

    try:
        site_fname = split_filename[0]
        species_fname = split_filename[1]
        file_sampling_period = split_filename[2]
        inlet_height = split_filename[3]
    except IndexError:
        raise ValueError(
            "Unable to read metadata from filename. We expect a filename such as tta.co2.1minute.222m.dat"
        )

    # metadata = read_metadata(filepath=data_filepath, data=data, data_type="ICOS")
    header = read_header(filepath=data_filepath)
    n_skip = len(header) - 1

    def date_parser(year: str, month: str, day: str, hour: str, minute: str) -> Timestamp:
        return Timestamp(year, month, day, hour, minute)

    datetime_columns = {"time": ["Year", "Month", "Day", "Hour", "Minute"]}

    use_cols = [
        "Year",
        "Month",
        "Day",
        "Hour",
        "Minute",
        str(species_fname.lower()),
        "Stdev",
        "NbPoints",
    ]

    dtypes = {
        "Day": int,
        "Month": int,
        "Year": int,
        "Hour": int,
        "Minute": int,
        species_fname.lower(): float,
        "Stdev": float,
        "SamplingHeight": float,
        "NbPoints": int,
    }

    data = read_csv(
        data_filepath,
        skiprows=n_skip,
        parse_dates=datetime_columns,
        index_col="time",
        sep=" ",
        usecols=use_cols,
        dtype=dtypes,
        na_values="-999.99",
        date_parser=date_parser,
    )

    data = data[data[species_fname.lower()] >= 0.0]

    # Drop duplicate indices
    data = data.loc[~data.index.duplicated(keep="first")]

    # Check if the index is sorted
    if not data.index.is_monotonic_increasing:
        data = data.sort_index()

    rename_dict = {
        "Stdev": species_fname + " variability",
        "NbPoints": species_fname + " number_of_observations",
    }

    data = data.rename(columns=rename_dict)

    # Convert to xarray Dataset
    data = data.to_xarray()

    if site_fname.lower() != site:
        raise ValueError("Site mismatch between site argument passed and filename")

    if inlet_height.lower() != inlet:
        raise ValueError("Mismatch between inlet height passed and in filename")

    if file_sampling_period == "1minute":
        file_sampling_period = "60.0"
    elif file_sampling_period == "1hour":
        file_sampling_period = "3600.0"

    if sampling_period is not None:
        if file_sampling_period != sampling_period:
            raise ValueError("Mismatch between sampling period read from filename and that passed in.")
    else:
        sampling_period = file_sampling_period

    metadata = {
        "site": site,
        "species": species_fname,
        "inlet": inlet_height,
        "sampling_period": sampling_period,
        "network": network,
        "instrument": instrument,
    }

    if measurement_type is not None:
        metadata["measurement_type"] = measurement_type

    species_data = {
        species_fname: {
            "metadata": metadata,
            "data": data,
        }
    }

    return species_data



from collections import abc
from dataclasses import dataclass
from json import dumps
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Iterator, Union

import plotly.graph_objects as go

from ._basedata import _BaseData

__all__ = ["ObsData"]


@dataclass(frozen=True)
class ObsData(_BaseData, abc.Mapping):
    """This class is used to return observations data from the get_observations function

    Args:
        data: Dictionary of xarray Dataframes
        metadata: Dictionary of metadata
    """

    # Compatability layer for legacy format - mimicking the behaviour of a dictionary
    # Previous format expected a dictionary containing the site code and data
    # as key:value pairs.
    # TODO: May also want to check other expected keys within the legacy
    # dictionary format and add them below
    def __getitem__(self, key: str) -> Any:
        """
        Returns the data attribute (xarray Dataset) when the site name is
        specified.
        Included as a compatability layer for legacy format as a dictionary
        containing a Dataset for each site code.

        key (str): Site code
        """
        site = self.metadata["site"].lower()
        if key.lower() == site:
            return self.data
        else:
            raise KeyError(f"Site '{key}' does not match to expected site '{site}'")

    def __iter__(self) -> Iterator:
        """
        Returns site code as the key for the dictionary as would be expected.
        """
        site = self.metadata["site"]
        return iter([site])

    def __len__(self) -> int:
        """
        Returns number of key values (fixed at 1 at present)
        """
        # Fixed length as 1 at the moment but may need to update if other key
        # values are added.
        return 1

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ObsData):
            return NotImplemented

        return self.data.equals(other.data) and self.metadata == other.metadata

    def to_data(self) -> Dict:
        """Creates a dictionary package of this ObsData's metadata and data.

        Returns:
            dict: Dictionary of metadata and bytes
        """
        to_transfer: Dict[str, Union[str, bytes]] = {}
        to_transfer["metadata"] = dumps(self.metadata)

        # TODO - get better bytes
        with NamedTemporaryFile() as tmpfile:
            self.data.to_netcdf(tmpfile.name)
            to_transfer["data"] = Path(tmpfile.name).read_bytes()

        return to_transfer

    def plot_timeseries(
        self, title: str = None, xlabel: str = None, ylabel: str = None, units: str = None
    ) -> go.Figure:
        """Plot a timeseries"""

        species = self.metadata["species"]
        site = self.metadata["site"]
        inlet = self.metadata["inlet"]

        if title is None:
            title = f"{species.upper()} at {site.upper()} - {inlet}"

        if xlabel is None:
            xlabel = "Date"

        if ylabel is None:
            ylabel = "Concentration"

        if units is not None:
            ylabel += f"  ({units})"

        data = self.data
        x_data = data.time

        try:
            y_data = data[species]
        except KeyError:
            y_data = data["mf"]

        font = {"size": 14}

        title_layout = {"text": title, "y": 0.9, "x": 0.5, "xanchor": "center", "yanchor": "top"}

        layout = go.Layout(
            title=title_layout,
            xaxis=dict(title=xlabel),
            yaxis=dict(title=ylabel),
            font=font,
        )

        # Create traces
        fig = go.Figure(layout=layout)
        fig.add_trace(go.Scatter(x=x_data, y=y_data, mode="lines", name=species.upper()))

        return fig



""" Handles the recombination of dataframes stored in the object store
    into the data requested by the user

"""
from typing import Dict, List, Optional, Union

import numpy as np
import xarray as xr
from xarray.core.coordinates import DatasetCoordinates

__all__ = ["recombine_multisite", "recombine_datasets"]


def recombine_multisite(keys: Dict, sort: Optional[bool] = True) -> Dict:
    """Recombine the keys from multiple sites into a single Dataset for each site

    Args:
        site_keys: A dictionary of lists of keys, keyed by site
        sort: Sort the resulting Dataset by the time dimension
    Returns:
        dict: Dictionary of xarray.Datasets
    """
    result = {}
    for key, key_list in keys.items():
        result[key] = recombine_datasets(keys=key_list, sort=sort)

    return result


def recombine_datasets(
    keys: List[str],
    sort: Optional[bool] = True,
    attrs_to_check: Dict[str, str] = None,
    elevate_inlet: bool = False,
) -> xr.Dataset:
    """Combines datasets stored separately in the object store
    into a single dataset

    Args:
        keys: List of object store keys
        sort: Sort the resulting Dataset by the time dimension. Default = True
        attrs_to_check: Attributes to check for duplicates. If duplicates are present
            a new data variable will be created containing the values from each dataset
            If a dictionary is passed, the attribute(s) will be retained and the new value assigned.
            If a list/string is passed, the attribute(s) will be removed.
        elevate_inlet: Force the elevation of the inlet attribute
    Returns:
        xarray.Dataset: Combined Dataset
    """
    from openghg.objectstore import get_bucket
    from openghg.store.base import Datasource
    from xarray import concat as xr_concat

    if not keys:
        raise ValueError("No data keys passed.")

    bucket = get_bucket()

    data = [Datasource.load_dataset(bucket=bucket, key=k) for k in keys]

    # Check if we've got multiple inlet heights
    inlets_to_check = check_inlets(data=data, elevate_inlet=elevate_inlet)

    if attrs_to_check is None:
        attrs_to_check = {}

    attrs_to_check.update(inlets_to_check)

    # For specified attributes (e.g. "inlet")
    # elevate duplicates to data variables within each Dataset
    if attrs_to_check:
        # if isinstance(attrs_to_check, dict):
        attributes = list(attrs_to_check.keys())
        replace_values = list(attrs_to_check.values())

        # TODO - GJ - 2022-02-22 - I'm not sure we need to many different ways of passing in inlets to check here?
        # elif isinstance(attrs_to_check, str):
        #     attributes = [attrs_to_check]
        #     replace_values = [""]
        # else:
        #     attributes = attrs_to_check
        #     replace_values = [""] * len(attributes)

        data = elevate_duplicate_attrs(ds_list=data, attributes=attributes, elevate_inlet=elevate_inlet)

    # Concatenate datasets along time dimension
    combined = xr_concat(data, dim="time")

    # Replace/remove incorrect attributes
    #  - xr.concat will only take value from first dataset if duplicated
    if attrs_to_check:
        for attr, value in zip(attributes, replace_values):
            if attr in combined:  # Only update if attr was elevated to a data variable
                if value:
                    combined.attrs[attr] = value
                else:
                    combined.attrs.pop(attr)

    if sort:
        combined = combined.sortby("time")

    # This is modified from https://stackoverflow.com/a/51077784/1303032
    unique, index, count = np.unique(combined.time, return_counts=True, return_index=True)
    n_dupes = unique[count > 1].size
    # dupes = unique[count > 1]

    if n_dupes > 5:
        raise ValueError("Large number of duplicate timestamps, check data overlap.")

    combined = combined.isel(time=index)

    return combined


def create_array_from_value(
    value: str,
    coords: Union[DatasetCoordinates, Dict[str, DatasetCoordinates]],  # type: ignore
    name: Union[str, None] = None,
) -> xr.DataArray:
    """
    Create a new xarray.DataArray object containing a single value repeated
    for each coordinate.

    Args:
        value: Value to be repeated within the DataArray object
        coords: Co-ordinates to use for this new DataArray.
        name: Name to give the variable within the DataArray
    Returns:
        DataArray
    """
    if isinstance(coords, xr.core.coordinates.DatasetCoordinates):
        names = list(coords.keys())
        dims = tuple(len(coords[n]) for n in names)
    elif isinstance(coords, dict):
        dims = tuple(len(coord) for coord in list(coords.values()))
    else:
        dims = (len(coords),)

    variable = np.tile(value, dims)
    data_variable = xr.DataArray(variable, coords=coords, name=name)

    return data_variable


def elevate_duplicate_attrs(
    ds_list: List[xr.Dataset], attributes: Union[str, List[str]], elevate_inlet: bool
) -> List[xr.Dataset]:
    """
    For a list of Datasets, if the specified attributes are being repeated
    these will be added as new data variables to each Dataset.

    Args:
        ds_list: List of xarray Datasets
        attributes: Attribute values to check within the Datasets. If None is passed
        the original dataset list will be returned.
        elevate_inlet: Force the elevation of inlet
    Returns:
        list: List of updated Dataset objects
    """
    if not isinstance(attributes, list):
        attributes = [attributes]

    for attr in attributes:
        # Pull the attributes out of the datasets - usually inlet values for ranked data
        data_attrs = [ds.attrs[attr] for ds in ds_list if attr in ds.attrs]

        # If we have more than one unique value we update the Dataset by adding a new variable
        # This is useful with ranked inlets so we can easily know which inlet a measurement was taken from
        if len(set(data_attrs)) > 1 or (attr == "inlet" and elevate_inlet):
            for i, ds in enumerate(ds_list):
                value = ds.attrs[attr]
                coords = ds.coords
                new_variable = create_array_from_value(value=value, coords=coords, name=attr)
                updated_ds = ds.assign({attr: new_variable})

                ds_list[i] = updated_ds

    return ds_list


def check_inlets(data: List[xr.Dataset], elevate_inlet: bool) -> Dict:
    """Check the inlets of the data to be processed

    Args:
        data: List of Datasets
    Returns:
        dict: Dictionary with single or multiple inlet replacement value
    """
    inlets = set()

    for dataset in data:
        try:
            inlets.add(dataset.attrs["inlet"])
        except KeyError:
            pass

    if len(inlets) > 1:
        attrs = {"inlet": "multiple"}
    else:
        if elevate_inlet:
            attrs = {"inlet": inlets.pop()}
        else:
            attrs = {}

    return attrs



from json import dump, loads
from pathlib import Path
from typing import Dict, List, Union

from addict import Dict as aDict
from openghg.dataobjects import ObsData

__all__ = ["to_dashboard", "to_dashboard_mobile"]


def to_dashboard(
    data: Union[ObsData, List[ObsData]],
    selected_vars: List,
    downsample_n: int = 3,
    filename: str = None,
) -> Union[Dict, None]:
    """Takes a Dataset produced by OpenGHG and outputs it into a JSON
    format readable by the OpenGHG dashboard or a related project.

    This also exports a separate file with the locations of the sites
    for use with map selector component.

    Note - this function does not currently support export of data from multiple
    inlets.

    Args:
        data: Dictionary of retrieved data
        selected_vars: The variables to want to export
        downsample_n: Take every nth value from the data
        filename: filename to write output to
    Returns:
        None
    """
    to_export = aDict()

    if not isinstance(selected_vars, list):
        selected_vars = [selected_vars]

    selected_vars = [str(c).lower() for c in selected_vars]

    if not isinstance(data, list):
        data = [data]

    for obs in data:
        measurement_data = obs.data
        attributes = measurement_data.attrs
        metadata = obs.metadata

        df = measurement_data.to_dataframe()

        rename_lower = {c: str(c).lower() for c in df.columns}
        df = df.rename(columns=rename_lower)
        # We just want the selected variables
        to_extract = [c for c in df.columns if c in selected_vars]

        if not to_extract:
            continue

        df = df[to_extract]

        # Downsample the data
        if downsample_n > 1:
            df = df.iloc[::downsample_n]

        network = metadata["network"]
        instrument = metadata["instrument"]

        try:
            latitude = attributes["latitude"]
        except KeyError:
            latitude = metadata["latitude"]

        try:
            longitude = attributes["longitude"]
        except KeyError:
            longitude = metadata["longitude"]

        # TODO - remove this if we add site location to standard metadata
        location = {
            "latitude": latitude,
            "longitude": longitude,
        }
        metadata.update(location)

        json_data = loads(df.to_json())

        species = metadata["species"]
        site = metadata["site"]
        inlet = metadata["inlet"]

        to_export[species][network][site][inlet][instrument] = {
            "data": json_data,
            "metadata": metadata,
        }

    if filename is not None:
        with open(filename, "w") as f:
            dump(obj=to_export, fp=f)
        return None
    else:
        # TODO - remove this once addict is stubbed
        export_dict: Dict = to_export.to_dict()
        return export_dict


def to_dashboard_mobile(data: Dict, filename: Union[str, Path] = None) -> Union[Dict, None]:
    """Export the Glasgow LICOR data to JSON for the dashboard

    Args:
        data: Data dictionary
        filename: Filename for export of JSON
    Returns:
        dict or None: Dictonary if no filename given
    """
    to_export = aDict()

    for species, species_data in data.items():
        spec_data = species_data["data"]
        metadata = species_data["metadata"]

        latitude = spec_data["latitude"].values.tolist()
        longitude = spec_data["longitude"].values.tolist()
        ch4 = spec_data["ch4"].values.tolist()

        to_export[species]["data"] = {"lat": latitude, "lon": longitude, "z": ch4}
        to_export[species]["metadata"] = metadata

    if filename is not None:
        with open(filename, "w") as f:
            dump(to_export, f)
        return None
    else:
        to_return: Dict = to_export.to_dict()
        return to_return
